"""Testing utilities for building HTTP-related expected queries.

Version Added:
    5.0.7
"""

from __future__ import annotations

from typing import Optional, TYPE_CHECKING, Union

from django.contrib.auth.models import AnonymousUser, User

from reviewboard.accounts.testing.queries import (
    get_user_by_pk_equeries,
    get_user_profile_equeries)
from reviewboard.site.testing.queries import \
    get_check_local_site_access_equeries

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries

    from reviewboard.site.models import LocalSite


def get_http_request_user_equeries(
    *,
    user: User,
) -> ExpectedQueries:
    """Return expected queries for fetching the user from the HTTP request.

    This corresponds to a call to :py:attr:`HttpRequest.user
    <django.http.HttpRequest.user>`.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.User):
            The user that's expected to be fetched.

    Returns:
        list of dict:
        The list of expected queries.
    """
    return get_user_by_pk_equeries(
        user=user,
        note=f'Fetch the logged-in user "{user.username}"')


def get_http_request_start_equeries(
    *,
    user: Union[AnonymousUser, User],
    local_site: Optional[LocalSite] = None,
    checks_local_site_access: bool = True,
) -> ExpectedQueries:
    """Return expected queries for the start of an HTTP request.

    This covers queries that fetch the user, their profile, and optionally
    the check for Local Site access.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user that's checked for access.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site value used in the queries.

        checks_local_site_access (bool, optional):
            Whether the requested view uses the
            :py:func:`~reviewboard.site.decorators.check_local_site_access`
            decorator.

    Returns:
        list of dict:
        The list of expected queries.
    """
    equeries: ExpectedQueries = []

    if user.is_authenticated:
        assert isinstance(user, User)

        equeries += get_http_request_user_equeries(user=user)
        equeries += get_user_profile_equeries(user=user)

    if local_site is not None and checks_local_site_access:
        equeries += get_check_local_site_access_equeries(user=user,
                                                         local_site=local_site)

    return equeries



"""Legacy imports for bug trackers."""

from __future__ import annotations

from housekeeping import ClassMovedMixin

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.hostingsvcs.base import bug_tracker


class BugTracker(ClassMovedMixin,
                 bug_tracker.BaseBugTracker,
                 warning_cls=RemovedInReviewBoard10_0Warning):
    """An interface to a bug tracker.

    Deprecated:
        8.0:
        This has been moved to :py:class:`reviewboard.hostingsvcs.base.
        bug_tracker.BaseBugTracker`. The legacy import will be removed in
        Review Board 10.
    """



"""Unit tests for reviewboard.reviews.models.Group"""

from django.contrib.auth.models import AnonymousUser
from djblets.testing.decorators import add_fixtures

from reviewboard.testing import TestCase


class ReviewGroupTests(TestCase):
    """Unit tests for reviewboard.reviews.models.Group."""

    def test_is_accessible_by_with_public(self):
        """Testing Group.is_accessible_by with public group"""
        user = self.create_user()
        group = self.create_review_group()

        self.assertTrue(group.is_accessible_by(user))
        self.assertTrue(group.is_accessible_by(AnonymousUser()))

    def test_is_accessible_by_with_public_and_hidden(self):
        """Testing Group.is_accessible_by with public hidden group"""
        user = self.create_user()
        group = self.create_review_group(visible=False)

        self.assertTrue(group.is_accessible_by(user))
        self.assertTrue(group.is_accessible_by(user))

    def test_is_accessible_by_with_invite_only_and_not_member(self):
        """Testing Group.is_accessible_by with invite-only group and user
        not a member
        """
        user = self.create_user()
        group = self.create_review_group(invite_only=True)

        self.assertFalse(group.is_accessible_by(user))
        self.assertFalse(group.is_accessible_by(AnonymousUser()))

    def test_is_accessible_by_with_invite_only_and_member(self):
        """Testing Group.is_accessible_by with invite-only group and user is
        a member
        """
        user = self.create_user()

        group = self.create_review_group(invite_only=True)
        group.users.add(user)

        self.assertTrue(group.is_accessible_by(user))

    def test_is_accessible_by_with_invite_only_and_superuser(self):
        """Testing Group.is_accessible_by with invite-only group and user is
        a superuser
        """
        user = self.create_user(is_superuser=True)
        group = self.create_review_group(invite_only=True)

        self.assertTrue(group.is_accessible_by(user))

    def test_is_accessible_by_with_invite_only_hidden_not_member(self):
        """Testing Group.is_accessible_by with invite-only hidden group and
        user not a member
        """
        user = self.create_user()
        group = self.create_review_group(invite_only=True,
                                         visible=False)

        self.assertFalse(group.is_accessible_by(user))
        self.assertFalse(group.is_accessible_by(AnonymousUser()))

    def test_is_accessible_by_with_invite_only_hidden_and_member(self):
        """Testing Group.is_accessible_by with invite-only hidden group and
        user is a member
        """
        user = self.create_user()

        group = self.create_review_group(invite_only=True,
                                         visible=False)
        group.users.add(user)

        self.assertTrue(group.is_accessible_by(user))

    @add_fixtures(['test_users', 'test_site'])
    def test_is_accessible_by_with_local_site_accessible(self):
        """Testing Group.is_accessible_by with Local Site accessible by user"""
        user = self.create_user()

        group = self.create_review_group(with_local_site=True)
        group.local_site.users.add(user)

        self.assertTrue(group.is_accessible_by(user))

    @add_fixtures(['test_users', 'test_site'])
    def test_is_accessible_by_with_local_site_not_accessible(self):
        """Testing Group.is_accessible_by with Local Site not accessible by
        user
        """
        user = self.create_user()
        group = self.create_review_group(with_local_site=True)

        self.assertFalse(group.is_accessible_by(user))
        self.assertFalse(group.is_accessible_by(AnonymousUser()))

    def test_is_mutable_by_with_non_admins(self):
        """Testing Group.is_mutable_by with non-administrative users"""
        user = self.create_user()
        group = self.create_review_group()

        self.assertFalse(group.is_mutable_by(user))
        self.assertFalse(group.is_mutable_by(AnonymousUser()))

    def test_is_mutable_by_with_superuser(self):
        """Testing Group.is_mutable_by with superuser"""
        user = self.create_user(is_superuser=True)
        group = self.create_review_group()

        self.assertTrue(group.is_mutable_by(user))

    def test_is_mutable_by_with_change_group_perm(self):
        """Testing Group.is_mutable_by with user with reviews.change_group
        permission
        """
        user = self.create_user(perms=[
            ('reviews', 'change_group'),
        ])
        group = self.create_review_group()

        self.assertTrue(group.is_mutable_by(user))

    @add_fixtures(['test_users', 'test_site'])
    def test_is_mutable_by_with_local_site_admin(self):
        """Testing Group.is_mutable_by with user with Local Site admin"""
        user = self.create_user()

        group = self.create_review_group(with_local_site=True)
        group.local_site.admins.add(user)

        self.assertTrue(group.is_mutable_by(user))



import json
from typing import Optional

from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.utils.translation import gettext as _
from djblets.secrets.token_generators import token_generator_registry
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_request_fields,
                                       webapi_response_errors)
from djblets.webapi.errors import (DOES_NOT_EXIST, INVALID_FORM_DATA,
                                   NOT_LOGGED_IN, PERMISSION_DENIED,
                                   WebAPITokenGenerationError)
from djblets.webapi.fields import (BooleanFieldType,
                                   DateTimeFieldType,
                                   DictFieldType,
                                   StringFieldType)

from reviewboard.webapi.base import ImportExtraDataError, WebAPIResource
from reviewboard.webapi.decorators import webapi_check_local_site
from reviewboard.webapi.errors import TOKEN_GENERATION_FAILED
from reviewboard.webapi.models import WebAPIToken
from reviewboard.webapi.resources import resources


class TokenExpiresFieldType(DateTimeFieldType):
    """A field type for the WebAPIToken's expires field.

    This field acts the same as
    :py:class:`djblets.webapi.fields.DateTimeFieldType` except that it can
    accept null values. This allows optional date/time fields such as the
    expires field to be set to date/time values or null.

    Version Added:
        5.0
    """

    name = _('Nullable ISO 8601 Date/Time')

    def clean_value(self, value):
        """Validate and return a null value or datetime from an ISO 8601 value.

        Args:
            value (object):
                The value to validate and normalize. This should be a
                :py:class:`datetime.datetime`, an ISO 8601 date/time string
                `None` or an empty string.

        Returns:
            datetime.datetime:
            The resulting date/time value.

        Raises:
            django.core.exceptions.ValidationError:
                The resulting value was not a null value, a valid ISO 8601
                date/time string or the time was ambiguous.
        """
        if value is None or value == '':
            return None

        return super(TokenExpiresFieldType, self).clean_value(value)


class APITokenResource(WebAPIResource):
    """Manages the tokens used to access the API.

    This resource allows callers to retrieve their list of tokens, register
    new tokens, delete old ones, and update information on existing tokens.
    """
    model = WebAPIToken
    name = 'api_token'
    verbose_name = 'API Token'

    api_token_access_allowed = False
    oauth2_token_access_allowed = False

    added_in = '2.5'

    fields = {
        'deprecated': {
            'type': BooleanFieldType,
            'description': 'Whether the token is deprecated. This is true if '
                           'the token was generated by a deprecated token '
                           'generator.',
            'added_in': '5.0.1',
        },
        'expires': {
            'type': TokenExpiresFieldType,
            'description': 'An optional field for the date and time that the '
                           'token will expire. The token will be invalid and '
                           'unusable for authentication after this point.',
            'added_in': '5.0',
        },
        'expired': {
            'type': BooleanFieldType,
            'description': 'Whether the token is expired.',
            'added_in': '5.0',
        },
        'extra_data': {
            'type': DictFieldType,
            'description': 'Extra data as part of the token. '
                           'This can be set by the API or extensions.',
        },
        'id': {
            'type': StringFieldType,
            'description': 'The numeric ID of the token entry.',
        },
        'invalid_date': {
            'type': DateTimeFieldType,
            'description': 'The date and time at which the token became '
                           'invalid.',
            'added_in': '5.0',
        },
        'invalid_reason': {
            'type': StringFieldType,
            'description': 'A message explaining why the token is no longer '
                           'valid.',
            'added_in': '5.0',
        },
        'last_updated': {
            'type': DateTimeFieldType,
            'description': 'The date and time that the token was last '
                           'updated.',
        },
        'last_used': {
            'type': DateTimeFieldType,
            'description': 'The date and time that the token was last '
                           'used for authentication.',
            'added_in': '5.0',
        },
        'note': {
            'type': StringFieldType,
            'description': 'The note explaining the purpose of this token.',
        },
        'policy': {
            'type': DictFieldType,
            'description': 'The access policies defined for this token.',
        },
        'time_added': {
            'type': DateTimeFieldType,
            'description': 'The date and time that the token was added.',
        },
        'token': {
            'type': StringFieldType,
            'description': 'The token value.',
        },
        'token_generator_id': {
            'type': StringFieldType,
            'description': 'The ID of the token generator that generated '
                           'the token.',
            'added_in': '5.0',
        },
        'valid': {
            'type': BooleanFieldType,
            'description': 'Whether the token is currently valid.',
            'added_in': '5.0',
        },
    }

    uri_object_key = 'api_token_id'
    last_modified_field = 'last_updated'
    model_parent_key = 'user'

    allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')

    def get_queryset(self, request, local_site_name=None, *args, **kwargs):
        user = resources.user.get_object(
            request, local_site_name=local_site_name, *args, **kwargs)

        local_site = self._get_local_site(local_site_name)

        return self.model.objects.filter(user=user, local_site=local_site)

    def has_list_access_permissions(self, request, *args, **kwargs):
        if request.user.is_superuser:
            return True

        user = resources.user.get_object(request, *args, **kwargs)
        return user == request.user

    def has_access_permissions(self, request, token, *args, **kwargs):
        return token.is_accessible_by(request.user)

    def has_modify_permissions(self, request, token, *args, **kwargs):
        return token.is_mutable_by(request.user)

    def has_delete_permissions(self, request, token, *args, **kwargs):
        return token.is_deletable_by(request.user)

    def serialize_deprecated_field(
        self,
        obj: WebAPIToken,
        *args,
        **kwargs,
    ) -> bool:
        """Serialize the ``deprecated`` field.

        Version Added:
            5.0.1

        Args:
            obj (reviewboard.webapi.models.WebAPIToken):
                The token that is being serialized.

            *args (tuple):
                Unused positional arguments.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            bool:
            Whether the API token is deprecated.
        """
        return obj.is_deprecated()

    def serialize_expired_field(
        self,
        obj: WebAPIToken,
        *args,
        **kwargs,
    ) -> bool:
        """Serialize the ``expired`` field.

        Version Added:
            5.0

        Args:
            obj (reviewboard.webapi.models.WebAPIToken):
                The token that is being serialized.

            *args (tuple):
                Unused positional arguments.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            bool:
            Whether the API token is expired.
        """
        return obj.is_expired()

    def serialize_expires_field(
        self,
        obj: WebAPIToken,
        *args,
        **kwargs,
    ) -> Optional[str]:
        """Serialize the ``expires`` field.

        Version Added:
            5.0

        Args:
            obj (reviewboard.webapi.models.WebAPIToken):
                The token that is being serialized.

            *args (tuple):
                Unused positional arguments.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            str or None:
            The expiration date of the token, in ISO-8601 format.
        """
        if obj.expires:
            return obj.expires.isoformat()
        else:
            return None

    def serialize_invalid_date_field(
        self,
        obj: WebAPIToken,
        *args,
        **kwargs,
    ) -> Optional[str]:
        """Serialize the ``invalid_date`` field.

        Version Added:
            5.0

        Args:
            obj (reviewboard.webapi.models.WebAPIToken):
                The token that is being serialized.

            *args (tuple):
                Unused positional arguments.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            str or None:
            The invalid date of the token, in ISO-8601 format.
        """
        if obj.invalid_date:
            return obj.invalid_date.isoformat()
        else:
            return None

    def serialize_last_updated_field(
        self,
        obj: WebAPIToken,
        *args,
        **kwargs,
    ) -> str:
        """Serialize the ``last_updated`` field.

        Version Added:
            5.0

        Args:
            obj (reviewboard.webapi.models.WebAPIToken):
                The token that is being serialized.

            *args (tuple):
                Unused positional arguments.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            str:
            The last updated date of the token, in ISO-8601 format.
        """
        return obj.last_updated.isoformat()

    def serialize_last_used_field(
        self,
        obj: WebAPIToken,
        *args,
        **kwargs,
    ) -> Optional[str]:
        """Serialize the ``last_used`` field.

        Version Added:
            5.0

        Args:
            obj (reviewboard.webapi.models.WebAPIToken):
                The token that is being serialized.

            *args (tuple):
                Unused positional arguments.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            str or None:
            The last used date of the token, in ISO-8601 format.
        """
        if obj.last_used:
            return obj.last_used.isoformat()
        else:
            return None

    def serialize_time_added_field(
        self,
        obj: WebAPIToken,
        *args,
        **kwargs,
    ) -> str:
        """Serialize the ``time_added`` field.

        Version Added:
            5.0

        Args:
            obj (reviewboard.webapi.models.WebAPIToken):
                The token that is being serialized.

            *args (tuple):
                Unused positional arguments.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            str:
            The time added date of the token, in ISO-8601 format.
        """
        return obj.time_added.isoformat()

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(DOES_NOT_EXIST, INVALID_FORM_DATA, NOT_LOGGED_IN,
                            PERMISSION_DENIED, TOKEN_GENERATION_FAILED)
    @webapi_request_fields(
        optional={
            'expires': {
                'type': TokenExpiresFieldType,
                'description': 'The date and time that the token will expire.'
                               'This must be a valid '
                               ':term:`date/time format`.',
                'added_in': '5.0',
            },
        },
        required={
            'note': {
                'type': StringFieldType,
                'description': 'The note explaining the purpose of '
                               'this token.',
            },
            'policy': {
                'type': StringFieldType,
                'description': 'The token access policy, encoded as a '
                               'JSON string.',
            },
        },
        allow_unknown=True
    )
    def create(self, request, note, policy, extra_fields={},
               local_site_name=None, *args, **kwargs):
        """Registers a new API token.

        The token value will be generated and returned in the payload.

        Callers are expected to provide a note and a policy.

        Note that this may, in theory, fail due to too many token collisions.
        If that happens, please re-try the request.

        Extra data can be stored later lookup. See
        :ref:`webapi2.0-extra-data` for more information.
        """
        try:
            user = resources.user.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        if not self.has_list_access_permissions(request, *args, **kwargs):
            return self.get_no_access_error(request)

        try:
            self._validate_policy(policy)
        except ValueError as e:
            return INVALID_FORM_DATA, {
                'fields': {
                    'policy': str(e),
                },
            }

        local_site = self._get_local_site(local_site_name)
        expires = kwargs.get('expires', None)
        token_generator_id = \
            token_generator_registry.get_default().token_generator_id

        try:
            token = WebAPIToken.objects.generate_token(
                user,
                token_generator_id=token_generator_id,
                token_info={'token_type': 'rbp'},
                note=note,
                policy=policy,
                expires=expires,
                local_site=local_site)
        except WebAPITokenGenerationError as e:
            return TOKEN_GENERATION_FAILED.with_message(str(e))

        if extra_fields:
            try:
                self.import_extra_data(token, token.extra_data, extra_fields)
            except ImportExtraDataError as e:
                return e.error_payload

            token.save(update_fields=('extra_data',))

        return 201, {
            self.item_result_key: token,
        }

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(DOES_NOT_EXIST, INVALID_FORM_DATA, NOT_LOGGED_IN,
                            PERMISSION_DENIED)
    @webapi_request_fields(
        optional={
            'expires': {
                'type': TokenExpiresFieldType,
                'description': 'The date and time that the token will expire.'
                               'This must be a valid '
                               ':term:`date/time format`.',
                'added_in': '5.0',
            },
            'invalid_reason': {
                'type': StringFieldType,
                'description': 'A message indicating why the token is '
                               'no longer valid.',
                'added_in': '5.0',
            },
            'note': {
                'type': StringFieldType,
                'description': 'The note explaining the purpose of '
                               'this token.',
            },
            'policy': {
                'type': StringFieldType,
                'description': 'The token access policy, encoded as a '
                               'JSON string.',
            },
            'valid': {
                'type': BooleanFieldType,
                'description': 'Whether the token is valid. This can only be '
                               'used to invalidate tokens by setting this to '
                               '``false``. Setting this to ``true`` is not '
                               'allowed and will cause an invalid form data '
                               'error.',
                'added_in': '5.0',
            },
        },
        allow_unknown=True
    )
    def update(self, request, extra_fields={}, *args, **kwargs):
        """Updates the information on an existing API token.

        The note, policy, and extra data on the token may be updated.
        See :ref:`webapi2.0-extra-data` for more information.

        This can also be used to invalidate a token by setting
        valid to ``false`` and including an invalid reason.
        """
        try:
            token = self.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        if not self.has_access_permissions(request, token, *args, **kwargs):
            return self.get_no_access_error(request)

        if 'expires' in kwargs:
            token.expires = kwargs['expires']
            token.extra_data['expired_notification_sent'] = False

        if 'note' in kwargs:
            token.note = kwargs['note']

        if 'policy' in kwargs:
            try:
                token.policy = self._validate_policy(kwargs['policy'])
            except ValidationError as e:
                return INVALID_FORM_DATA, {
                    'fields': {
                        'policy': e.message,
                    },
                }

        valid = kwargs.get('valid')

        if valid is True:
            return INVALID_FORM_DATA, {
                'fields': {
                    'valid': _('This can only be used to invalidate the '
                               'token. You cannot set valid to true.')
                }
            }
        elif valid is False:
            self.model.objects.invalidate_token(
                token=token.token,
                invalid_reason=kwargs.get('invalid_reason', ''))

        if extra_fields:
            try:
                self.import_extra_data(token, token.extra_data, extra_fields)
            except ImportExtraDataError as e:
                return e.error_payload

        token.save()

        return 200, {
            self.item_result_key: token,
        }

    @augment_method_from(WebAPIResource)
    def delete(self, *args, **kwargs):
        """Delete the API token, invalidating all clients using it.

        The API token will be removed from the user's account, and will no
        longer be usable for authentication.

        After deletion, this will return a :http:`204`.
        """
        pass

    @augment_method_from(WebAPIResource)
    def get_list(self, *args, **kwargs):
        """Retrieves a list of API tokens belonging to a user.

        If accessing this API on a Local Site, the results will be limited
        to those associated with that site.

        This can only be accessed by the owner of the tokens, or superusers.
        """
        pass

    @augment_method_from(WebAPIResource)
    def get(self, *args, **kwargs):
        """Retrieves information on a particular API token.

        This can only be accessed by the owner of the tokens, or superusers.
        """
        pass

    def _validate_policy(self, policy_str):
        try:
            policy = json.loads(policy_str)
        except Exception as e:
            raise ValidationError(
                _('The policy is not valid JSON: %s')
                % str(e))

        self.model.validate_policy(policy)

        return policy


api_token_resource = APITokenResource()



"""Unit tests for reviewboard.webapi.resources.session."""

from __future__ import annotations

import json
from typing import TYPE_CHECKING

from django.http import SimpleCookie
from djblets.webapi.errors import NOT_LOGGED_IN
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import session_mimetype
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.urls import get_session_url

if TYPE_CHECKING:
    from typing import Any

    from django.contrib.auth.models import User
    from typelets.json import JSONDict

    from reviewboard.webapi.tests.mixins import BasicPutTestSetupState


class ResourceTests(BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Testing the SessionResource APIs."""

    fixtures = ['test_users']
    sample_api_url = 'session/'
    resource = resources.session

    def setup_http_not_allowed_list_test(self, user):
        return get_session_url()

    def setup_http_not_allowed_item_test(self, user):
        return get_session_url()

    def compare_item(self, item_rsp, user):
        self.assertTrue(item_rsp['authenticated'])
        self.assertEqual(item_rsp['links']['user']['title'], user.username)
        self.assertEqual(item_rsp['links']['delete']['href'],
                         item_rsp['links']['self']['href'])
        self.assertEqual(item_rsp['links']['delete']['method'], 'DELETE')

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        return (get_session_url(local_site_name),
                session_mimetype,
                user)

    def test_get_with_anonymous_user(self):
        """Testing the GET session/ API with anonymous user"""
        self.client.logout()

        rsp = self.api_get(get_session_url(),
                           expected_mimetype=session_mimetype)

        self.assertEqual(
            rsp,
            {
                'session': {
                    'authenticated': False,
                    'links': {
                        'self': {
                            'href': 'http://testserver/api/session/',
                            'method': 'GET',
                        },
                    },
                },
                'stat': 'ok',
            })

    #
    # HTTP DELETE test
    #

    def setup_basic_delete_test(self, user, with_local_site, local_site_name):
        return (get_session_url(local_site_name),
                session_mimetype)

    def check_delete_result(self, user, *args):
        pass

    @webapi_test_template
    def test_delete_not_owner(self):
        """Testing the DELETE <URL> API when not logged in"""
        self.load_fixtures(self.basic_delete_fixtures)

        url, cb_args = self.setup_basic_delete_test(self.user, False, None)

        self.client.logout()
        self.client.cookies = SimpleCookie()

        rsp = self.api_delete(url, expected_status=401)

        self.assertEqual(
            rsp,
            {
                'err': {
                    'code': NOT_LOGGED_IN.code,
                    'msg': NOT_LOGGED_IN.msg,
                    'type': NOT_LOGGED_IN.error_type,
                },
                'stat': 'fail',
            })

    #
    # HTTP PUT tests
    #

    # This test does not apply, so remove it.
    test_put_not_owner = None

    def populate_put_test_objects(
        self,
        *,
        setup_state: BasicPutTestSetupState,
        create_valid_request_data: bool,
        **kwargs,
    ) -> None:
        """Populate objects for a PUT test.

        Version Added:
            8.0

        Args:
            setup_state (reviewboard.webapi.tests.mixins.
                         BasicPutTestSetupState):
                The setup state for the test.

            create_valid_request_data (bool):
                Whether ``request_data`` in ``setup_state`` should provide
                valid data for a PUT test, given the populated objects.

            **kwargs (dict):
                Additional keyword arguments for future expansion.
        """
        setup_state.update({
            'item': None,
            'mimetype': session_mimetype,
            'request_data': {},
            'url': get_session_url(
                local_site_name=setup_state['local_site_name']),
        })

    def check_put_result(
        self,
        user: User,
        item_rsp: JSONDict,
        item: Any,
        *args,
    ) -> None:
        """Check the results of an HTTP PUT.

        Version Added:
            8.0

        Args:
            user (django.contrib.auth.models.User):
                The user performing the requesdt.

            item_rsp (dict):
                The item payload from the response.

            item (object):
                The item to compare to.

            *args (tuple):
                Positional arguments provided by
                :py:meth:`setup_basic_put_test`.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        self.compare_item(item_rsp, user)

    @webapi_test_template
    def test_put_with_settings_json(self) -> None:
        """Testing the PUT <URL> API with settings:json"""
        user = self._login_user()

        profile = user.get_profile()
        profile.settings['some_setting'] = '123'
        profile.save(update_fields=('settings',))

        rsp = self.api_put(
            get_session_url(),
            {
                'settings:json': json.dumps({
                    'bad_option': '123',
                    'confirm_ship_it': False,
                })
            },
            expected_mimetype=session_mimetype)

        self.assertEqual(
            rsp,
            {
                'session': {
                    'authenticated': True,
                    'links': {
                        'delete': {
                            'href': 'http://testserver/api/session/',
                            'method': 'DELETE',
                        },
                        'self': {
                            'href': 'http://testserver/api/session/',
                            'method': 'GET',
                        },
                        'user': {
                            'href': 'http://testserver/api/users/grumpy/',
                            'method': 'GET',
                            'title': 'grumpy',
                        },
                    },
                },
                'stat': 'ok',
            })

        profile.refresh_from_db()
        self.assertEqual(profile.settings, {
            'some_setting': '123',
            'confirm_ship_it': False,
        })



"""Testing utilities for building expected queries for accounts.

Version Added:
    5.0.7
"""

from __future__ import annotations

from typing import Optional, TYPE_CHECKING

from django.contrib.auth.models import Permission, User
from django.db.models import Q

from reviewboard.accounts.models import LocalSiteProfile, Profile

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries

    from reviewboard.site.models import LocalSite


def get_user_by_pk_equeries(
    *,
    user: User,
    note: Optional[str] = None,
) -> ExpectedQueries:
    """Return expected queries for fetching a user by ID.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user that's expected to be fetched.

        note (str, optional):
            An optional note to use, instead of the default.

    Returns:
        list of dict:
        The list of expected queries.
    """
    return [
        {
            '__note__': note or f'Fetch the user {user.username}',
            'model': User,
            'where': Q(pk=user.pk),
        },
    ]


def get_user_local_site_profile_equeries(
    *,
    user: User,
    profile: Profile,
    local_site: Optional[LocalSite] = None,
) -> ExpectedQueries:
    """Return expected queries for accessing a user's LocalSiteProfile.

    This corresponds to a call to :py:meth:`User.get_site_profile()`.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user with the profile.

        profile (reviewboard.accounts.models.Profile):
            The user's primary profile.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site associated with the profile.

    Returns:
        list of dict:
        The list of expected queries.
    """
    return [
        {
            '__note__': "Fetch the user's LocalSiteProfile",
            'model': LocalSiteProfile,
            'where': (Q(local_site=local_site) &
                      Q(profile=profile) &
                      Q(user=user)),
        }
    ]


def get_user_permissions_equeries(
    *,
    user: User,
) -> ExpectedQueries:
    """Return expected queries for fetching a user's permissions.

    This corresponds to a call to :py:meth:`StandardAuthBackend.has_perm()
    <reviewboard.accounts.backends.standard.StandardAuthBackend>`, without
    passing a Local Site.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user with the profile.

    Returns:
        list of dict:
        The list of expected queries.
    """
    username = user.username

    return [
        {
            '__note__': f'Fetch the user "{username}"\'s permissions',
            'join_types': {
                'auth_user_user_permissions': 'INNER JOIN',
                'django_content_type': 'INNER JOIN',
            },
            'model': Permission,
            'num_joins': 2,
            'tables': {
                'auth_permission',
                'auth_user_user_permissions',
                'django_content_type',
            },
            'values_select': ('content_type__app_label', 'codename'),
            'where': Q(user__id=user.pk),
        },
        {
            '__note__': f'Fetch the user "{username}"\'s permission groups',
            'join_types': {
                'auth_group': 'INNER JOIN',
                'auth_group_permissions': 'INNER JOIN',
                'auth_user_groups': 'INNER JOIN',
                'django_content_type': 'INNER JOIN',
            },
            'model': Permission,
            'num_joins': 4,
            'tables': {
                'auth_group',
                'auth_group_permissions',
                'auth_permission',
                'auth_user_groups',
                'django_content_type',
            },
            'values_select': ('content_type__app_label', 'codename'),
            'where': Q(group__user=user),
        },
    ]


def get_user_profile_equeries(
    *,
    user: User,
) -> ExpectedQueries:
    """Return expected queries for accessing a user's profile.

    This corresponds to a call to :py:meth:`User.get_profile()`.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user with the profile.

    Returns:
        list of dict:
        The list of expected queries.
    """
    return [
        {
            '__note__': "Fetch the user's profile",
            'model': Profile,
            'where': Q(user=user),
        },
    ]



"""Hooks for adding info the the dashboard."""

from __future__ import annotations

from djblets.extensions.hooks import (DataGridColumnsHook,
                                      ExtensionHook,
                                      ExtensionHookPoint)

from reviewboard.datagrids.grids import (DashboardDataGrid,
                                         UserPageReviewRequestDataGrid)


class DataGridSidebarItemsHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """A hook for adding items to the sidebar of a datagrid.

    Extensions can use this hook to plug new items into the sidebar of
    any datagrid supporting sidebars.

    The items can be any subclass of
    :py:class:`reviewboard.datagrids.sidebar.BaseSidebarItem`, including the
    built-in :py:class:`reviewboard.datagrids.sidebar.BaseSidebarSection` and
    built-in :py:class:`reviewboard.datagrids.sidebar.SidebarNavItem`.
    """

    def initialize(self, datagrid, item_classes):
        """Initialize the hook.

        This will register the provided datagrid sidebar item classes in the
        provided datagrid.

        Args:
            datagrid (type):
                The datagrid class to register the items on. The datagrid
                must have a sidebar, or an error will occur.

            item_classes (list of type):
                The list of item classes to register on the datagrid's
                sidebar. Each must be a subclass of
                :py:class:`~reviewboard.datagrids.sidebar.BaseSidebarItem`.

        Raises:
            ValueError:
                A datagrid was provided that does not contain a sidebar.
        """
        if not hasattr(datagrid, 'sidebar'):
            raise ValueError('The datagrid provided does not have a sidebar')

        self.datagrid = datagrid
        self.item_classes = item_classes

        for item in item_classes:
            datagrid.sidebar.add_item(item)

    def shutdown(self):
        """Shut down the hook.

        This will unregister each item class from the datagrid's sidebar.
        """
        for item in self.item_classes:
            self.datagrid.sidebar.remove_item(item)


# We don't use the ExtensionHookPoint metaclass here, because we actually
# want these to register in the base DataGridColumnsHook point.
class DashboardColumnsHook(DataGridColumnsHook):
    """A hook for adding custom columns to the dashboard.

    Extensions can use this hook to provide one or more custom columns
    in the dashboard. These columns can be added by users, moved around,
    and even sorted, like other columns.

    Each value passed to ``columns`` must be an instance of
    :py:class:`djblets.datagrid.grids.Column`.

    It also must have an ``id`` attribute set. This must be unique within
    the dashboard. It is recommended to use a vendor-specific prefix to the
    ID, in order to avoid conflicts.
    """

    def initialize(self, columns):
        """Initialize the hook.

        This will register each of the provided columns on the Dashboard.

        Args:
            columns (list of djblets.datagrid.grids.Column):
                The list of column instances to register on the Dashboard.
        """
        super(DashboardColumnsHook, self).initialize(DashboardDataGrid,
                                                     columns)


class DashboardSidebarItemsHook(DataGridSidebarItemsHook,
                                metaclass=ExtensionHookPoint):
    """A hook for adding items to the sidebar of the dashboard.

    Extensions can use this hook to plug new items into the sidebar of
    the dashboard. These will appear below the built-in items.

    The items can be any subclass of
    :py:class:`reviewboard.datagrids.sidebar.BaseSidebarItem`, including the
    built-in :py:class:`reviewboard.datagrids.sidebar.BaseSidebarSection` and
    built-in :py:class:`reviewboard.datagrids.sidebar.SidebarNavItem`.
    """

    def initialize(self, item_classes):
        """Initialize the hook.

        This will register the provided datagrid sidebar item classes in the
        Dashboard.

        Args:
            item_classes (list of type):
                The list of item classes to register on the datagrid's
                sidebar. Each must be a subclass of
                :py:class:`~reviewboard.datagrids.sidebar.BaseSidebarItem`.
        """
        super(DashboardSidebarItemsHook, self).initialize(DashboardDataGrid,
                                                          item_classes)


class UserPageSidebarItemsHook(DataGridSidebarItemsHook,
                               metaclass=ExtensionHookPoint):
    """A hook for adding items to the sidebar of the user page.

    Extensions can use this hook to plug new items into the sidebar of
    the user page. These will appear below the built-in items.

    The items can be any subclass of
    :py:class:`reviewboard.datagrids.sidebar.BaseSidebarItem`, including the
    built-in :py:class:`reviewboard.datagrids.sidebar.BaseSidebarSection` and
    built-in :py:class:`reviewboard.datagrids.sidebar.SidebarNavItem`.
    """

    def initialize(self, item_classes):
        """Initialize the hook.

        This will register the provided datagrid sidebar item classes in the
        user page's datagrid.

        Args:
            item_classes (list of type):
                The list of item classes to register on the datagrid's
                sidebar. Each must be a subclass of
                :py:class:`~reviewboard.datagrids.sidebar.BaseSidebarItem`.
        """
        super(UserPageSidebarItemsHook, self).initialize(
            UserPageReviewRequestDataGrid, item_classes)



"""Tests for reviewboard.diffviewer.managers.DiffCommitManager."""

import kgb
from dateutil.parser import parse as parse_date
from django.db import IntegrityError

from reviewboard.diffviewer.models import DiffCommit, DiffSet
from reviewboard.diffviewer.parser import DiffParser
from reviewboard.testing.testcase import TestCase


class DiffCommitManagerTests(kgb.SpyAgency, TestCase):
    """Unit tests for DiffCommitManager."""

    fixtures = ['test_scmtools']

    commit_test_diff = (
        b'diff --git a/README b/README\n'
        b'index d6613f5..5b50866 100644\n'
        b'--- README\n'
        b'+++ README\n'
        b'@ -1,1 +1,1 @@\n'
        b'-blah..\n'
        b'+blah blah\n'
    )

    def test_create_from_data(self):
        """Testing DiffCommitManager.create_from_data"""
        repository = self.create_repository(tool_name='Test')
        self.spy_on(repository.get_file_exists,
                    op=kgb.SpyOpReturn(True))

        diffset = DiffSet.objects.create_empty(
            repository=repository,
            basedir='',
            revision=1)

        raw_date = '2000-01-01 00:00:00-0600'
        parsed_date = parse_date(raw_date)
        commit = DiffCommit.objects.create_from_data(
            repository=repository,
            diff_file_name='diff',
            diff_file_contents=self.commit_test_diff,
            parent_diff_file_name=None,
            parent_diff_file_contents=b'',
            request=None,
            commit_id='r1',
            parent_id='r0',
            author_name='Author',
            author_email='author@example.com',
            author_date=parsed_date,
            committer_name='Committer',
            committer_email='committer@example.com',
            committer_date=parsed_date,
            commit_message='Description',
            diffset=diffset,
            validation_info={})

        self.assertEqual(commit.files.count(), 1)
        self.assertEqual(diffset.files.count(), commit.files.count())
        self.assertEqual(diffset.commit_count, 1)

        # We have to compare regular equality and equality after applying
        # ``strftime`` because two datetimes with different timezone info
        # may be equal
        self.assertEqual(commit.author_date, parsed_date)
        self.assertEqual(commit.committer_date, parsed_date)
        self.assertEqual(commit.committer_date_utc, parsed_date)
        self.assertEqual(commit.committer_date_offset, -21600.0)

        self.assertEqual(
            commit.author_date.strftime(DiffCommit.ISO_DATE_FORMAT),
            raw_date)

        self.assertEqual(
            commit.committer_date.strftime(DiffCommit.ISO_DATE_FORMAT),
            raw_date)

    def test_create_from_data_with_author_date_none(self):
        """Testing DiffCommitManager.create_from_data with author_date=None"""
        # author_date is a property that sets a couple of fields based on
        # the provided date. If this crashes (which was an issue initially with
        # the implementation when passing in None), construction will fail
        # with:
        #
        #     TypeError: 'author_date' is an invalid keyword argument for this
        #     function
        #
        # Instead, we set to None and allow Django to raise a standard
        # IntegrityError on model save.
        repository = self.create_repository(tool_name='Test')
        self.spy_on(repository.get_file_exists,
                    op=kgb.SpyOpReturn(True))

        diffset = DiffSet.objects.create_empty(
            repository=repository,
            basedir='',
            revision=1)

        with self.assertRaises(IntegrityError):
            DiffCommit.objects.create_from_data(
                repository=repository,
                diff_file_name='diff',
                diff_file_contents=self.commit_test_diff,
                parent_diff_file_name=None,
                parent_diff_file_contents=b'',
                request=None,
                commit_id='r1',
                parent_id='r0',
                author_name='Author',
                author_email='author@example.com',
                author_date=None,
                committer_name='Committer',
                committer_email='committer@example.com',
                committer_date=parse_date('2000-01-01 00:00:00-0600'),
                commit_message='Description',
                diffset=diffset,
                validation_info={})

        self.assertEqual(diffset.commit_count, 0)

    def test_create_from_data_with_committer_date_none(self):
        """Testing DiffCommitManager.create_from_data with committer_date=None
        """
        # committer_date is a property that sets a couple of fields based on
        # the provided date. If this crashes (which was an issue initially with
        # the implementation when passing in None), construction will fail
        # with:
        #
        #     TypeError: 'committer_date' is an invalid keyword argument for
        #     this function
        repository = self.create_repository(tool_name='Test')
        self.spy_on(repository.get_file_exists,
                    op=kgb.SpyOpReturn(True))

        diffset = DiffSet.objects.create_empty(
            repository=repository,
            basedir='',
            revision=1)

        commit = DiffCommit.objects.create_from_data(
            repository=repository,
            diff_file_name='diff',
            diff_file_contents=self.commit_test_diff,
            parent_diff_file_name=None,
            parent_diff_file_contents=b'',
            request=None,
            commit_id='r1',
            parent_id='r0',
            author_name='Author',
            author_email='author@example.com',
            author_date=parse_date('2000-01-01 00:00:00-0600'),
            committer_name='Committer',
            committer_email='committer@example.com',
            committer_date=None,
            commit_message='Description',
            diffset=diffset,
            validation_info={})

        self.assertEqual(commit.files.count(), 1)
        self.assertEqual(diffset.files.count(), commit.files.count())
        self.assertEqual(diffset.commit_count, 1)

        self.assertIsNone(commit.committer_date)
        self.assertIsNone(commit.committer_date_utc)
        self.assertIsNone(commit.committer_date_offset)

    def test_create_from_data_custom_parser_extra_data(self):
        """Testing DiffCommitManager.create_from_data with a custom diff parser
        that sets extra_data
        """
        repository = self.create_repository(tool_name='Test')

        class CustomParser(DiffParser):
            def parse(self):
                result = super(CustomParser, self).parse()

                self.parsed_diff.extra_data = {
                    'key1': 'value1',
                }

                self.parsed_diff_change.extra_data = {
                    'key2': 'value2',
                }

                return result

            def parse_diff_header(self, linenum, parsed_file):
                parsed_file.extra_data = {
                    'key3': 'value3',
                }

                return super(CustomParser, self).parse_diff_header(
                    linenum, parsed_file)

        self.spy_on(repository.get_file_exists,
                    op=kgb.SpyOpReturn(True))

        tool = repository.get_scmtool()

        self.spy_on(repository.get_scmtool,
                    op=kgb.SpyOpReturn(tool))
        self.spy_on(tool.get_parser,
                    call_fake=lambda repo, diff: CustomParser(diff))

        diffset = DiffSet.objects.create_empty(
            repository=repository,
            basedir='',
            revision=1)

        diffcommit = DiffCommit.objects.create_from_data(
            repository=repository,
            diff_file_name='diff',
            diff_file_contents=self.DEFAULT_FILEDIFF_DATA_DIFF,
            parent_diff_file_name=None,
            parent_diff_file_contents=b'',
            request=None,
            commit_id='r1',
            parent_id='r0',
            author_name='Author',
            author_email='author@example.com',
            author_date=parse_date('2000-01-01 00:00:00-0600'),
            committer_name='Committer',
            committer_email='committer@example.com',
            committer_date=None,
            commit_message='Description',
            diffset=diffset,
            validation_info={})

        # Test against what's in the database.
        diffset.refresh_from_db()
        diffcommit.refresh_from_db()

        self.assertEqual(diffset.extra_data, {
            'key1': 'value1',
        })

        self.assertEqual(diffcommit.extra_data, {
            'key2': 'value2',
        })

        self.assertEqual(diffset.files.count(), 1)

        filediff = diffset.files.all()[0]
        self.assertEqual(filediff.extra_data, {
            'is_symlink': False,
            'key3': 'value3',
            'raw_delete_count': 1,
            'raw_insert_count': 1,
        })

        # Check the FileLookupContext passed to get_file_exists.
        self.assertSpyCallCount(repository.get_file_exists, 1)

        context = repository.get_file_exists.last_call.kwargs.get('context')
        self.assertIsNotNone(context)
        self.assertEqual(context.diff_extra_data, {
            'key1': 'value1',
        })
        self.assertEqual(context.commit_extra_data, {
            'key2': 'value2',
        })
        self.assertEqual(context.file_extra_data, {
            'key3': 'value3',
        })



"""Built-in sidebar items for datagrids."""

from __future__ import annotations

from typing import Iterator, Optional, Sequence, TYPE_CHECKING

from django.utils.translation import gettext_lazy as _

from reviewboard.datagrids.sidebar import (BaseSidebarItem,
                                           BaseSidebarSection, SidebarNavItem)
from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from django.db.models import QuerySet

    from reviewboard.reviews.models import Group


class OverviewSection(BaseSidebarItem):
    """The "Overview" section on the Dashboard sidebar.

    This links to a Dashboard view showing all open incoming review requests
    listing the user directly as a reviewer, or listing a group the user is
    a member of and all open outgoing review requests made by the user.
    """

    template_name = 'datagrids/overview.html'
    label = _('Overview')
    view_id = 'overview'


class OutgoingSection(BaseSidebarSection):
    """The "Outgoing" section on the Dashboard sidebar.

    This displays two links: "All" and "Open".

    "All" links to a Dashboard view showing all outgoing review requests
    made by the user, including those that are closed.

    "Open" links to a Dashboard view showing only open outgoing review
    requests made by the user.
    """

    label = _('Outgoing')

    def get_items(self):
        """Yield each of the items within this section."""
        site_profile = self.datagrid.site_profile

        yield SidebarNavItem(self,
                             label=_('All'),
                             view_id='mine',
                             count=site_profile.total_outgoing_request_count)
        yield SidebarNavItem(self,
                             label=_('Open'),
                             view_id='outgoing',
                             count=site_profile.pending_outgoing_request_count)


class IncomingSection(BaseSidebarSection):
    """The "Incoming" section on the Dashboard sidebar.

    This displays three special links ("Open", "To Me", and "Starred"), and
    links for every group the user is a member of or has starred.

    "Open" links to a Dashboard view showing all open incoming review
    requests listing the user directly as a reviewer, or listing a group the
    user is a member of.

    "To Me" links to a Dashboard view showing all open incoming review
    requests listing the user directly as a reviewer.

    "Starred" links to a Dashboard view showing all review requests the
    user has starred. This will only show up if the user has any starred.

    Each group links to a Dashboard view showing all open review requests
    made to that group.
    """

    label = _('Incoming')

    def get_items(self) -> Iterator[SidebarNavItem]:
        """Yield each of the items within this section.

        Yields:
            reviewboard.datagrids.sidebar.SidebarNavItem:
            Each sidebar navigation item.
        """
        datagrid = self.datagrid
        profile = datagrid.profile
        local_site = datagrid.local_site
        site_profile = datagrid.site_profile

        yield SidebarNavItem(self,
                             label=_('Open'),
                             view_id='incoming',
                             count=site_profile.total_incoming_request_count)

        yield SidebarNavItem(self,
                             label=_('To Me'),
                             view_id='to-me',
                             count=site_profile.direct_incoming_request_count)

        if site_profile.starred_public_request_count > 0:
            yield SidebarNavItem(
                self,
                label=_('Starred'),
                view_id='starred',
                icon_name='rb-icon-star-on',
                count=site_profile.starred_public_request_count)

        only_fields = ('incoming_request_count', 'name', 'pk')

        groups = _get_groups(
            queryset=(
                datagrid.user.review_groups
                .only(*only_fields)
            ),
            local_site=local_site)

        seen_groups = {
            group.pk
            for group in groups
        }

        yield from self._add_groups(groups, view_id='to-group')

        if profile.has_starred_review_groups(local_site=local_site):
            yield from self._add_groups(
                _get_groups(
                    queryset=(
                        profile.starred_groups
                        .exclude(pk__in=seen_groups)
                        .only(*only_fields)
                    ),
                    local_site=local_site),
                view_id='to-watched-group',
                icon_name='rb-icon-star-on')

    def _add_groups(
        self,
        groups: Sequence[Group],
        view_id: str,
        icon_name: Optional[str] = None,
    ) -> Iterator[SidebarNavItem]:
        """Add groups to the sidebar.

        This will generate a sidebar navigation item for each group in the
        list.

        Args:
            groups (list of reviewboard.reviews.models.group.Group):
                The list of groups to add.

            view_id (str):
                The ID of the view to link to.

            icon_name (str, optional):
                The name of the icon to show.

        Yields:
            reviewboard.datagrids.sidebar.SidebarNavItem:
            Each group navigation item.
        """
        for i, group in enumerate(groups):
            name = group.name

            if i == 0:
                css_classes = ['new-subsection']
            else:
                css_classes = []

            yield SidebarNavItem(self,
                                 label=name,
                                 view_id=view_id,
                                 view_args={
                                     'group': name,
                                 },
                                 icon_name=icon_name,
                                 css_classes=css_classes,
                                 count=group.incoming_request_count)


class UserProfileItem(BaseSidebarItem):
    """Displays the profile for a user in the user page sidebar.

    This will display information such as the name, e-mail address,
    gravatar, and dates logged in and joined.
    """

    template_name = 'datagrids/sidebar_user_info.html'

    def get_extra_context(self):
        """Return extra data to include in the template render context."""
        request = self.datagrid.request
        user = self.datagrid.user

        return {
            'show_profile': user.is_profile_visible(request.user),
            'profile_user': user,
        }


class UserGroupsItem(BaseSidebarSection):
    """Displays the list of groups a user belongs to in the user page sidebar.

    Each group will be clickable, and will navigate to the corresponding
    group page.
    """

    label = _('Groups')
    css_classes = ['-is-desktop-only']

    def get_items(self) -> Iterator[SidebarNavItem]:
        """Yield each of the items within this section.

        Yields:
            reviewboard.datagrids.sidebar.SidebarNavItem:
            Each sidebar navigation item.
        """
        datagrid = self.datagrid

        groups = _get_groups(
            queryset=(
                datagrid.user.review_groups
                .accessible(datagrid.request.user)
                .only('local_site', 'name', 'pk')
            ),
            local_site=datagrid.local_site)

        for group in groups:
            yield SidebarNavItem(self,
                                 label=group.name,
                                 url=group.get_absolute_url())


def _get_groups(
    *,
    queryset: QuerySet[Group],
    local_site: Optional[LocalSite],
) -> Sequence[Group]:
    """Return review groups for usage in a datagrid sidebar.

    This takes a queryset, conditionally applies a Local Site filter if
    needed, and applies a sort Python-side, returning the list of results.

    Callers should take care to provide a queryset that performs any
    accessibility checks if needed, and to return the minimum number of
    fields used for the item in the sidebar (which must always include "name").

    Version Added:
        5.0.7

    Args:
        queryset (django.db.models.QuerySet):
            The queryset for review groups.

        local_site (reviewboard.site.models.LocalSite):
            The Local Site being viewed, if any.

    Returns:
        list of reviewboard.reviews.models.group.Group:
        The resulting list of sorted review groups.
    """
    local_site_q = LocalSite.objects.build_q(local_site, allow_all=False)

    if local_site_q:
        queryset = queryset.filter(local_site_q)

    return sorted(queryset,
                  key=lambda group: group.name)



"""Error definitions for Forgejo.

Version Added:
    8.0
"""

from __future__ import annotations


class APITokenNameInUseError(Exception):
    """Error for when an API token name is already in use.

    Version Added:
        8.0
    """



"""Module for interacting with different source code management tools."""

from djblets.registries.importer import lazy_import_registry

from reviewboard.signals import initializing


scmtools_registry = lazy_import_registry(
    'reviewboard.scmtools.registry',
    'SCMToolRegistry')


def _populate_registry(**kwargs):
    """Populate the SCMTools registry.

    Version Added:
        5.0

    Args:
        **kwargs (dict, unused):
            Keyword arguments sent by the signal.
    """
    # TODO: We do this here because we want the registry to be populated early
    # in case we need to create any new entries for the Tool table. We can't do
    # it in AppConfig.ready because that's not allowed to do database queries
    # (and can break things like building docs as a result). Once we get rid of
    # the Tool model, we can either move this into an AppConfig, or just let
    # the registry populate itself the first time it's accessed.
    scmtools_registry.populate()


initializing.connect(_populate_registry)



"""Administration form for SSH settings."""

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext, gettext_lazy as _

from reviewboard.ssh.client import SSHClient


class SSHSettingsForm(forms.Form):
    """SSH key settings for Review Board."""

    keyfile = forms.FileField(
        label=_('Key file'),
        required=False,
        widget=forms.FileInput(attrs={'size': '35'}))

    # These will ultimately map to the submit buttons.
    generate_key = forms.BooleanField(
        required=False,
        initial=True,
        widget=forms.HiddenInput)
    delete_key = forms.BooleanField(
        required=False,
        initial=True,
        widget=forms.HiddenInput)
    upload_key = forms.BooleanField(
        required=False,
        initial=True,
        widget=forms.HiddenInput)

    def clean(self):
        """Clean the form data.

        This will perform an additional validation check to see if the user
        requested to upload a key, but failed to provide a key file.

        Returns:
            dict:
            The resulting cleaned data for the form.
        """
        cleaned_data = super(SSHSettingsForm, self).clean()

        if cleaned_data.get('upload_key') and not cleaned_data.get('keyfile'):
            self.add_error(
                'keyfile',
                ValidationError(
                    self.fields['keyfile'].error_messages['required'],
                    code='required'))

        return cleaned_data

    def create(self, files):
        """Generate or import an SSH key.

        This will generate a new SSH key if :py:attr:`generate_key` was set
        to ``True``. Otherwise, a if :py:attr:`keyfile` was provided, its
        corresponding file upload will be used as the new key.

        In either case, the key will be validated, and if validation fails,
        an error will be set for the appropriate field.

        Args:
            files (django.utils.datastructures.MultiValueDict):
                The files uploaded in the request. This may contain a
                ``keyfile`` entry representing a key to upload.

        Raises:
            Exception:
                There was an error generating or importing a key. The form
                will have a suitable error for the field triggering the
                error.
        """
        if self.cleaned_data['generate_key']:
            try:
                SSHClient().generate_user_key()
            except IOError as e:
                self.add_error(
                    'generate_key',
                    gettext('Unable to write SSH key file: %s') % e)
                raise
            except Exception as e:
                self.add_error(
                    'generate_key',
                    gettext('Error generating SSH key: %s') % e)
                raise
        elif self.cleaned_data['upload_key']:
            try:
                SSHClient().import_user_key(files['keyfile'])
            except IOError as e:
                self.add_error(
                    'keyfile',
                    gettext('Unable to write SSH key file: %s') % e)
                raise
            except Exception as e:
                self.add_error(
                    'keyfile',
                    gettext('Error uploading SSH key: %s') % e)
                raise

    def did_request_delete(self):
        """Return whether the user has requested to delete the user SSH key.

        This checks that :py:attr:`delete_key`` was set in the request.

        Returns:
            ``True`` if the user requested to delete the key. ``False`` if
            they did not.
        """
        return self.cleaned_data.get('delete_key', False)

    def delete(self):
        """Delete the configured SSH user key.

        This will only delete the key if :py:attr:`delete_key` was set.

        Raises:
            Exception:
                There was an unexpected error deleting the key. A validation
                error will be set for the ``delete_key`` field.
        """
        if self.did_request_delete():
            try:
                SSHClient().delete_user_key()
            except Exception as e:
                self.add_error(
                    'delete_key',
                    gettext('Unable to delete SSH key file: %s') % e)
                raise

    class Meta:
        title = _('SSH Settings')



"""Unit tests for reviewboard.certs.utils.

Version Added:
    8.0
"""

from __future__ import annotations

from reviewboard.certs.utils import (get_cert_hostname_matches,
                                     normalize_cert_hostname)
from reviewboard.testing import TestCase


class GetCertHostnameMatchesTests(TestCase):
    """Unit tests for get_cert_hostname_matches.

    Version Added:
        8.0
    """

    def test_with_equal(self) -> None:
        """Testing get_cert_hostname_matches with normalized hostnames equal"""
        self.assertTrue(get_cert_hostname_matches(
            cert_hostname='www.example.COM',
            check_hostname='WWW.EXAMPLE.COM',
        ))
        self.assertTrue(get_cert_hostname_matches(
            cert_hostname='www.example.COM',
            check_hostname='www.example.COM',
            normalize_hostnames=False,
        ))

    def test_with_not_equal(self) -> None:
        """Testing get_cert_hostname_matches with hostnames not equal and
        non-wildcard
        """
        self.assertFalse(get_cert_hostname_matches(
            cert_hostname='example.com',
            check_hostname='www.example.com',
        ))
        self.assertFalse(get_cert_hostname_matches(
            cert_hostname='www.example.com',
            check_hostname='www.example.COM',
            normalize_hostnames=False,
        ))

    def test_with_wildcard(self) -> None:
        """Testing get_cert_hostname_matches with wildcard for full first
        label
        """
        # These should match.
        self.assertTrue(get_cert_hostname_matches(
            cert_hostname='*.example.com',
            check_hostname='www.example.com',
        ))
        self.assertTrue(get_cert_hostname_matches(
            cert_hostname='*.example.com',
            check_hostname='foobar.example.com',
        ))
        self.assertTrue(get_cert_hostname_matches(
            cert_hostname='*.example.com',
            check_hostname='FOOBAR.EXAMPLE.COM',
        ))

        # These should not.
        self.assertFalse(get_cert_hostname_matches(
            cert_hostname='*.example.com',
            check_hostname='example.com',
        ))
        self.assertFalse(get_cert_hostname_matches(
            cert_hostname='*.example.com',
            check_hostname='sub.www.example.com',
        ))
        self.assertFalse(get_cert_hostname_matches(
            cert_hostname='*.example.com',
            check_hostname='FOOBAR.EXAMPLE.COM',
            normalize_hostnames=False,
        ))


class NormalizeCertHostnameTests(TestCase):
    """Unit tests for normalize_cert_hostname.

    Version Added:
        8.0
    """

    def test_with_casing(self) -> None:
        """Testing normalize_cert_hostname with casing differences"""
        self.assertEqual(normalize_cert_hostname('Example.COM'),
                         'example.com')



"""Unit tests for the HostingServiceAccountResource."""

from __future__ import annotations

from email.message import Message
from http import HTTPStatus
from typing import Any, Optional, Sequence, TYPE_CHECKING
from urllib.error import HTTPError

import kgb
from djblets.webapi.errors import INVALID_FORM_DATA
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.hostingsvcs.base.client import HostingServiceClient
from reviewboard.hostingsvcs.base.http import (HostingServiceHTTPRequest,
                                               HostingServiceHTTPResponse)
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.webapi.errors import HOSTINGSVC_AUTH_ERROR
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (
    hosting_service_account_item_mimetype,
    hosting_service_account_list_mimetype)
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.urls import (
    get_hosting_service_account_item_url,
    get_hosting_service_account_list_url)

if TYPE_CHECKING:
    from django.contrib.auth.models import User
    from typelets.json import JSONDict


class ResourceListTests(kgb.SpyAgency,
                        BaseWebAPITestCase,
                        metaclass=BasicTestsMetaclass):
    """Unit tests for the HostingServiceAccountResource list APIs."""

    sample_api_url = 'hosting-services-accounts/'
    resource = resources.hosting_service_account
    fixtures = ['test_users']

    basic_post_use_admin = True

    def compare_item(
        self,
        item_rsp: JSONDict,
        account: HostingServiceAccount,
    ) -> None:
        """Compare an API response to an item.

        Args:
            item_rsp (dict):
                The API response.

            account (reviewboard.hostingsvcs.models.HostingServiceAccount):
                The account object to compare to.
        """
        self.assertEqual(item_rsp['id'], account.pk)
        self.assertEqual(item_rsp['username'], account.username)
        self.assertEqual(item_rsp['service'],
                         account.service.hosting_service_id)

    def setup_http_not_allowed_item_test(
        self,
        user: User,
    ) -> str:
        """Set up the HTTP not allowed test for item access.

        Args:
            user (django.contrib.auth.models.User):
                The user performing the API requests.

        Returns:
            str:
            The URL to use for accessing the resource.
        """
        return get_hosting_service_account_list_url()

    def setup_http_not_allowed_list_test(
        self,
        user: User,
    ) -> str:
        """Set up the HTTP not allowed test for list access.

        Args:
            user (django.contrib.auth.models.User):
                The user performing the API requests.

        Returns:
            str:
            The URL to use for accessing the resource.
        """
        return get_hosting_service_account_list_url()

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(
        self,
        user: User,
        with_local_site: bool,
        local_site_name: Optional[str],
        populate_items: bool,
    ) -> tuple[str, str, list[HostingServiceAccount]]:
        """Set up a basic HTTP GET test.

        Args:
            user (django.contrib.auth.models.User):
                The user performing the API requests.

            with_local_site (bool):
                Whether to run the tests against a Local Site.

            local_site_name (str):
                The name of the Local Site to use.

            populate_items (bool):
                Whether to populate items in the database.

        Returns:
            tuple:
            A 3-tuple of:

            Tuple:
                0 (str):
                    The URL to use for accessing the resource.

                1 (str):
                    The expected mimetype for responses.

                2 (list):
                    A list of accounts that are expected to be in the response.
        """
        if populate_items:
            if with_local_site:
                assert local_site_name is not None
                local_site = self.get_local_site(name=local_site_name)
            else:
                local_site = None

            accounts = [
                HostingServiceAccount.objects.create(
                    service_name='googlecode',
                    username='bob',
                    local_site=local_site),
            ]
        else:
            accounts = []

        return (get_hosting_service_account_list_url(local_site_name),
                hosting_service_account_list_mimetype,
                accounts)

    @webapi_test_template
    def test_get_with_service(self) -> None:
        """Testing the GET <URL>?service= API"""
        HostingServiceAccount.objects.create(
            service_name='googlecode',
            username='bob')

        account = HostingServiceAccount.objects.create(
            service_name='github',
            username='bob')

        rsp = self.api_get(
            get_hosting_service_account_list_url(),
            data={'service': 'github'},
            expected_mimetype=hosting_service_account_list_mimetype)
        assert rsp is not None

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(len(rsp['hosting_service_accounts']), 1)
        self.compare_item(rsp['hosting_service_accounts'][0], account)

    @webapi_test_template
    def test_get_with_username(self) -> None:
        """Testing the GET <URL>?username= API"""
        account = HostingServiceAccount.objects.create(
            service_name='googlecode',
            username='bob')

        HostingServiceAccount.objects.create(
            service_name='googlecode',
            username='frank')

        rsp = self.api_get(
            get_hosting_service_account_list_url(),
            data={'username': 'bob'},
            expected_mimetype=hosting_service_account_list_mimetype)
        assert rsp is not None

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(len(rsp['hosting_service_accounts']), 1)
        self.compare_item(rsp['hosting_service_accounts'][0], account)

    #
    # HTTP POST tests
    #

    def setup_basic_post_test(
        self,
        user: User,
        with_local_site: bool,
        local_site_name: Optional[str],
        post_valid_data: bool,
    ) -> tuple[str, str, JSONDict, Sequence[Any]]:
        """Set up a basic HTTP POST test.

        Args:
            user (django.contrib.auth.models.User):
                The user performing the API requests.

            with_local_site (bool):
                Whether to run the tests against a Local Site.

            local_site_name (str):
                The name of the Local Site to use.

            post_valid_data (bool):
                Whether to use valid data for the POST body.

        Returns:
            tuple:
            A 4-tuple of:

            Tuple:
                0 (str):
                    The URL to use for accessing the resource.

                1 (str):
                    The expected mimetype of the response.

                2 (dict):
                    The data to send in the POST request.

                3 (list):
                    Additional positional arguments to pass to
                    :py:meth:`check_post_result`.
        """
        if post_valid_data:
            post_data: JSONDict = {
                'username': 'bob',
                'service_id': 'googlecode',
            }
        else:
            post_data = {}

        return (get_hosting_service_account_list_url(local_site_name),
                hosting_service_account_item_mimetype,
                post_data,
                [])

    def check_post_result(
        self,
        user: User,
        rsp: JSONDict,
    ) -> None:
        """Check the result of a POST operation.

        Args:
            user (django.contrib.auth.models.User):
                The user performing the API requests.

            rsp (dict):
                The response content.
        """
        item_rsp = rsp['hosting_service_account']
        account = HostingServiceAccount.objects.get(pk=item_rsp['id'])

        self.compare_item(item_rsp, account)

    @webapi_test_template
    def test_post_with_gitlab_authorization(self) -> None:
        """Testing the POST <URL> API with a GitLab account"""
        url = 'http://example.com'
        data = b''

        request = HostingServiceHTTPRequest(url)
        response = HostingServiceHTTPResponse(
            request, url, data, {}, HTTPStatus.OK)

        self.spy_on(HostingServiceClient.http_get,
                    owner=HostingServiceClient,
                    op=kgb.SpyOpReturn(response))

        self._login_user(local_site=False, admin=True)

        rsp = self.api_post(
            get_hosting_service_account_list_url(),
            {
                'hosting_url': 'https://gitlab.com/',
                'private_token': 'mytoken',
                'service_id': 'gitlab',
                'username': 'test',
            },
            expected_status=201,
            expected_mimetype=hosting_service_account_item_mimetype)
        assert rsp is not None

        self.assertEqual(rsp['stat'], 'ok')
        self.assertSpyCalledWith(
            HostingServiceClient.http_get,
            'https://gitlab.com/api/v4/projects?per_page=1',
            headers={'PRIVATE-TOKEN': 'mytoken'})

        item_rsp = rsp['hosting_service_account']
        account = HostingServiceAccount.objects.get(pk=item_rsp['id'])
        self.compare_item(item_rsp, account)

    @webapi_test_template
    def test_post_with_gitlab_authorization_unauthorized(self) -> None:
        """Testing the POST <URL> API with a GitLab account and authorization
        failure
        """
        error = HTTPError(
            'http://example.com',
            HTTPStatus.UNAUTHORIZED,
            '',
            Message(),
            None)

        self.spy_on(HostingServiceClient.http_get,
                    owner=HostingServiceClient,
                    op=kgb.SpyOpRaise(error))

        self._login_user(local_site=False, admin=True)

        rsp = self.api_post(
            get_hosting_service_account_list_url(),
            {
                'hosting_url': 'https://gitlab.com/',
                'private_token': 'mytoken',
                'service_id': 'gitlab',
                'username': 'test',
            },
            expected_status=403)
        assert rsp is not None

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], HOSTINGSVC_AUTH_ERROR.code)
        self.assertEqual(rsp['reason'], 'The API token is invalid.')

    @webapi_test_template
    def test_post_with_gitlab_missing_fields(self) -> None:
        """Testing the POST <URL> API with a GitLab account and missing fields
        in custom HostingServiceAuthForm
        """
        self._login_user(local_site=False, admin=True)

        rsp = self.api_post(
            get_hosting_service_account_list_url(),
            {
                'hosting_url': 'https://gitlab.com/',
                'service_id': 'gitlab',
                'username': 'test',
            },
            expected_status=400)
        assert rsp is not None

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertEqual(rsp['fields'], {
            'private_token': ['This field is required.'],
        })


class ResourceItemTests(BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Unit tests for the HostingServiceAccountResource item APIs."""

    fixtures = ['test_users']
    sample_api_url = 'hosting-service-accounts/<id>/'
    resource = resources.hosting_service_account

    def compare_item(
        self,
        item_rsp: JSONDict,
        account: HostingServiceAccount,
    ) -> None:
        """Compare an API response to an item.

        Args:
            item_rsp (dict):
                The API response.

            account (reviewboard.hostingsvcs.models.HostingServiceAccount):
                The account object to compare to.
        """
        self.assertEqual(item_rsp['id'], account.pk)
        self.assertEqual(item_rsp['username'], account.username)
        self.assertEqual(item_rsp['service'],
                         account.service.hosting_service_id)

    def setup_http_not_allowed_item_test(
        self,
        user: User,
    ) -> str:
        """Set up the HTTP not allowed test for item access.

        Args:
            user (django.contrib.auth.models.User):
                The user performing the API requests.

        Returns:
            str:
            The URL to use for accessing the resource.
        """
        account = HostingServiceAccount.objects.create(
            service_name='googlecode',
            username='bob')

        return get_hosting_service_account_item_url(account.pk)

    def setup_http_not_allowed_list_test(
        self,
        user: User,
    ) -> str:
        """Set up the HTTP not allowed test for list access.

        Args:
            user (django.contrib.auth.models.User):
                The user performing the API requests.

        Returns:
            str:
            The URL to use for accessing the resource.
        """
        account = HostingServiceAccount.objects.create(
            service_name='googlecode',
            username='bob')

        return get_hosting_service_account_item_url(account.pk)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(
        self,
        user: User,
        with_local_site: bool,
        local_site_name: Optional[str],
    ) -> tuple[str, str, HostingServiceAccount]:
        """Set up a basic HTTP GET test.

        Args:
            user (django.contrib.auth.models.User):
                The user performing the API requests.

            with_local_site (bool):
                Whether to run the tests against a Local Site.

            local_site_name (str):
                The name of the Local Site to use.

        Returns:
            tuple:
            A 3-tuple of:

            Tuple:
                0 (str):
                    The URL to use for making API requests.

                1 (str):
                    The expected mimetype of the response.

                2 (reviewboard.hostingsvcs.models.HostingServiceAccount):
                    The hosting service account.
        """
        if with_local_site:
            assert local_site_name is not None
            local_site = self.get_local_site(name=local_site_name)
        else:
            local_site = None

        account = HostingServiceAccount.objects.create(
            service_name='googlecode',
            username='bob',
            local_site=local_site)

        return (get_hosting_service_account_item_url(account, local_site_name),
                hosting_service_account_item_mimetype,
                account)



"""Unit tests for reviewboard.accounts.models.LocalSiteProfile."""

import kgb
from django.contrib.auth.models import User

from reviewboard.accounts.models import LocalSiteProfile
from reviewboard.testing import TestCase


class LocalSiteProfileTests(kgb.SpyAgency, TestCase):
    """Unit tests for reviewboard.accounts.models.LocalSiteProfile."""

    fixtures = ['test_users']

    def test_duplicate_profiles(self):
        """Testing LocalSiteProfileManager.for_user consolidation of duplicate
        profiles
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile(create_if_missing=True)

        self.create_review_request(public=True, submitter=user)

        self.spy_on(LocalSiteProfile.objects._fix_duplicate_profiles)

        LocalSiteProfile.objects.create(user=user, profile=profile)
        LocalSiteProfile.objects.create(user=user, profile=profile)
        LocalSiteProfile.objects.create(
            user=user,
            profile=profile,
            permissions={'test': True})

        # Make sure we don't have any cached profiles.
        user._site_profiles = {}

        profile = user.get_site_profile(local_site=None)

        # Check merged profile data.
        self.assertEqual(profile.permissions, {'test': True})
        self.assertIsNone(profile.direct_incoming_request_count)
        self.assertIsNone(profile.total_incoming_request_count)
        self.assertIsNone(profile.pending_outgoing_request_count)
        self.assertIsNone(profile.total_outgoing_request_count)
        self.assertIsNone(profile.starred_public_request_count)

        # And make sure all those changes are persisted.
        profile.refresh_from_db()
        self.assertEqual(profile.permissions, {'test': True})
        self.assertEqual(profile.direct_incoming_request_count, 0)
        self.assertEqual(profile.total_incoming_request_count, 0)
        self.assertEqual(profile.pending_outgoing_request_count, 1)
        self.assertEqual(profile.total_outgoing_request_count, 1)
        self.assertEqual(profile.starred_public_request_count, 0)

        self.assertSpyCalled(LocalSiteProfile.objects._fix_duplicate_profiles)



"""Unit tests for Hosting Service registration."""

from __future__ import annotations

from typing import TYPE_CHECKING

from django.http import HttpResponse
from django.urls import NoReverseMatch, path
from djblets.registries.errors import AlreadyRegisteredError, ItemLookupError

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService
from reviewboard.hostingsvcs.base.registry import hosting_service_registry
from reviewboard.hostingsvcs.service import (register_hosting_service,
                                             unregister_hosting_service)
from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from django.http import HttpRequest


def hosting_service_url_test_view(
    request: HttpRequest,
    repo_id: int,
) -> HttpResponse:
    """View to test URL pattern addition when registering a hosting service.

    Args:
        request (django.http.HttpRequest):
            The HTTP request.

        repo_id (int):
            The ID of the repository.

    Returns:
        django.http.HttpResponse:
        A fake response.
    """
    return HttpResponse(str(repo_id))


class HostingServiceRegistrationTests(TestCase):
    """Unit tests for Hosting Service registration."""

    class _DummyService(BaseHostingService):
        name = 'DummyService'
        hosting_service_id = 'dummy-service'

    class _DummyServiceWithURLs(BaseHostingService):
        name = 'DummyServiceWithURLs'
        hosting_service_id = 'dummy-service'

        repository_url_patterns = [
            path('hooks/pre-commit/',
                 hosting_service_url_test_view,
                 name='dummy-service-post-commit-hook'),
        ]

    def tearDown(self) -> None:
        """Tear down the test case."""
        super().tearDown()

        # Unregister the service, going back to a default state. It's okay
        # if it fails.
        #
        # This will match whichever service we added for testing.
        try:
            hosting_service_registry.unregister_by_id('dummy-service')
        except ItemLookupError:
            pass

    def test_register_without_urls(self) -> None:
        """Testing HostingService registration"""
        hosting_service_registry.register(self._DummyService)

        with self.assertRaises(AlreadyRegisteredError):
            hosting_service_registry.register(self._DummyService)

    def test_unregister(self) -> None:
        """Testing HostingService unregistration"""
        hosting_service_registry.register(self._DummyService)
        hosting_service_registry.unregister(self._DummyService)

    def test_registration_with_urls(self) -> None:
        """Testing HostingService registration with URLs"""
        hosting_service_registry.register(self._DummyServiceWithURLs)

        self.assertEqual(
            local_site_reverse(
                'dummy-service-post-commit-hook',
                kwargs={
                    'repository_id': 1,
                    'hosting_service_id': 'dummy-service',
                }),
            '/repos/1/dummy-service/hooks/pre-commit/')

        self.assertEqual(
            local_site_reverse(
                'dummy-service-post-commit-hook',
                local_site_name='test-site',
                kwargs={
                    'repository_id': 1,
                    'hosting_service_id': 'dummy-service',
                }),
            '/s/test-site/repos/1/dummy-service/hooks/pre-commit/')

        # Once registered, should not be able to register again.
        with self.assertRaises(AlreadyRegisteredError):
            hosting_service_registry.register(self._DummyServiceWithURLs)

    def test_unregistration_with_urls(self) -> None:
        """Testing HostingService unregistration with URLs"""
        hosting_service_registry.register(self._DummyServiceWithURLs)
        hosting_service_registry.unregister(self._DummyServiceWithURLs)

        with self.assertRaises(NoReverseMatch):
            local_site_reverse(
                'dummy-service-post-commit-hook',
                kwargs={
                    'repository_id': 1,
                    'hosting_service_id': 'dummy-service',
                })

        # Once unregistered, should not be able to unregister again.
        with self.assertRaises(ItemLookupError):
            hosting_service_registry.unregister(self._DummyServiceWithURLs)

    def test_legacy_register_methods(self) -> None:
        """Testing legacy register and unregister methods"""
        message = (
            '`reviewboard.hostingsvcs.service.register_hosting_service()` '
            'has moved to `reviewboard.hostingsvcs.base.registry.'
            'HostingServiceRegistry.register`. The old function is '
            'deprecated and will be removed in Review Board 10.0.'
        )

        with self.assertWarns(RemovedInReviewBoard10_0Warning, message):
            register_hosting_service('dummy-service', self._DummyService)

        message = (
            '`reviewboard.hostingsvcs.service.unregister_hosting_service()` '
            'has moved to `reviewboard.hostingsvcs.base.registry.'
            'HostingServiceRegistry.unregister_by_id`. The old function is '
            'deprecated and will be removed in Review Board 10.0.'
        )

        with self.assertWarns(RemovedInReviewBoard10_0Warning, message):
            unregister_hosting_service('dummy-service')



"""Management command to condense stored diffs in the database."""

import sys
from datetime import datetime, timedelta

from django.conf import settings
from django.contrib.humanize.templatetags.humanize import intcomma
from django.core.management.base import BaseCommand
from django.utils.translation import gettext as _, ngettext_lazy as N_

from reviewboard.diffviewer.models import FileDiff


class Command(BaseCommand):
    """Management command to condense stored diffs in the database."""

    help = _('Condenses the diffs stored in the database, reducing space '
             'requirements')

    DELAY_SHOW_REMAINING_SECS = 30

    TIME_REMAINING_CHUNKS = (
        (60 * 60 * 24 * 365, N_('%d year', '%d years')),
        (60 * 60 * 24 * 30, N_('%d month', '%d months')),
        (60 * 60 * 24 * 7, N_('%d week', '%d weeks')),
        (60 * 60 * 24, N_('%d day', '%d days')),
        (60 * 60, N_('%d hour', '%d hours')),
        (60, N_('%d minute', '%d minutes'))
    )

    # We add a bunch of spaces in order to override any previous
    # content on the line, for when it shrinks.
    TIME_REMAINING_STR = _('%s remaining                                  ')

    CALC_TIME_REMAINING_STR = _('Calculating time remaining')

    def add_arguments(self, parser):
        """Add arguments to the command.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser for the command.
        """
        parser.add_argument(
            '--show-counts-only',
            action='store_true',
            dest='show_counts',
            default=False,
            help=_("Show the number of diffs that are expected to be "
                   "migrated, but don't perform a migration."))
        parser.add_argument(
            '--no-progress',
            action='store_false',
            dest='show_progress',
            default=True,
            help=_("Don't show progress information or totals while "
                   "migrating. You might want to use this if your database "
                   "is taking too long to generate total migration counts."))
        parser.add_argument(
            '--max-diffs',
            action='store',
            dest='max_diffs',
            type=int,
            default=None,
            help=_("The maximum number of migrations to perform. This is "
                   "useful if you have a lot of diffs to migrate and want "
                   "to do it over several sessions."))

    def handle(self, **options):
        """Handle the command.

        Args:
            **options (dict, unused):
                Options parsed on the command line.

        Raises:
            django.core.management.CommandError:
                There was an error performing a diff migration.
        """
        self.show_progress = options['show_progress']
        max_diffs = options['max_diffs']

        if options['show_counts']:
            counts = FileDiff.objects.get_migration_counts()
            self.stdout.write(_('%d unmigrated Review Board pre-1.7 diffs\n')
                              % counts['filediffs'])
            self.stdout.write(_('%d unmigrated Review Board 1.7-2.5 diffs\n')
                              % counts['legacy_file_diff_data'])
            self.stdout.write(_('%d total unmigrated diffs\n')
                              % counts['total_count'])
            return
        elif self.show_progress:
            counts = FileDiff.objects.get_migration_counts()
            total_count = counts['total_count']

            if total_count == 0:
                self.stdout.write(_('All diffs have already been migrated.\n'))
                return

            warning = counts.get('warning')

            if warning:
                self.stderr.write(_('Warning: %s\n\n') % warning)

            if max_diffs is None:
                process_count = total_count
            else:
                process_count = min(max_diffs, total_count)

            self.stdout.write(_('Processing %(count)d unmigrated diffs...\n')
                              % {'count': process_count})
        else:
            # Set to an empty dictionary to force migrate_all() to not
            # look up its own counts.
            counts = {}

            self.stdout.write(_('Processing all unmigrated diffs...\n'))

        self.stdout.write(_(
          '\n'
          'This may take a while. It is safe to continue using '
          'Review Board while this is\n'
          'processing, but it may temporarily run slower.\n'
          '\n'))

        # Don't allow queries to be stored.
        settings.DEBUG = False

        self.start_time = datetime.now()
        self.prev_prefix_len = 0
        self.prev_time_remaining_s = ''
        self.show_remaining = False

        info = FileDiff.objects.migrate_all(batch_done_cb=self._on_batch_done,
                                            counts=counts,
                                            max_diffs=max_diffs)

        if info['diffs_migrated'] == 0:
            self.stdout.write(_('All diffs have already been migrated.\n'))
        else:
            old_diff_size = info['old_diff_size']
            new_diff_size = info['new_diff_size']

            self.stdout.write(
                _('\n'
                  '\n'
                  'Condensed stored diffs from %(old_size)s bytes to '
                  '%(new_size)s bytes (%(savings_pct)0.2f%% savings)\n')
                % {
                    'old_size': intcomma(old_diff_size),
                    'new_size': intcomma(new_diff_size),
                    'savings_pct': (float(old_diff_size - new_diff_size) /
                                    float(old_diff_size) * 100),
                })

    def _on_batch_done(self, total_diffs_migrated, total_count=None, **kwargs):
        """Handler for when a batch of diffs are processed.

        This will report the progress of the operation, showing the estimated
        amount of time remaining.

        Args:
            total_diffs_migrated (int):
                The total number of diffs migrated so far in this
                condensediffs operation.

            total_count (int, optional):
                The total number of diffs to migrate in the database. This
                may be ``None``, in which case the output won't contain
                progress and time estimation.

            **kwargs (dict, unused):
                Unused keyword arguments.
        """
        # NOTE: We use sys.stdout when writing instead of self.stderr in order
        #       to control newlines. Command.stderr will force a \n for each
        #       write.
        if self.show_progress:
            # We may be receiving an estimate for the total number of diffs
            # that is less than the actual count. If we've gone past the
            # initial total, just bump up the total to the current count.
            total_count = max(total_diffs_migrated, total_count)

            pct = total_diffs_migrated * 100 / total_count

            delta = datetime.now() - self.start_time
            delta_secs = delta.total_seconds()

            if (not self.show_remaining and
                delta_secs >= self.DELAY_SHOW_REMAINING_SECS):
                self.show_remaining = True

            if self.show_remaining:
                secs_left = ((delta_secs // total_diffs_migrated) *
                             (total_count - total_diffs_migrated))

                time_remaining_s = (self.TIME_REMAINING_STR
                                    % self._time_remaining(secs_left))
            else:
                time_remaining_s = self.CALC_TIME_REMAINING_STR

            prefix_s = '  [%d%%] %s/%s - ' % (pct, total_diffs_migrated,
                                              total_count)

            sys.stdout.write(prefix_s)

            # Only write out the time remaining string if it has changed or
            # there's been a shift in the length of the prefix. This reduces
            # how much we have to write to the terminal, and how often, by
            # a fair amount.
            if (self.prev_prefix_len != len(prefix_s) or
                self.prev_time_remaining_s != time_remaining_s):
                # Something has changed, so output the string and then cache
                # the values for the next call.
                sys.stdout.write(time_remaining_s)

                self.prev_prefix_len = len(prefix_s)
                self.prev_time_remaining_s = time_remaining_s
        else:
            sys.stdout.write(' %s diffs migrated' % total_diffs_migrated)

        sys.stdout.write('\r')
        sys.stdout.flush()

    def _time_remaining(self, secs_left):
        """Return a string representing the time remaining for the operation.

        This is a simplified version of Django's timeuntil() function that
        does fewer calculations in order to reduce the amount of time we
        have to spend every loop. For instance, it doesn't bother with
        constructing datetimes and recomputing deltas, since we already
        have those, and it doesn't rebuild the TIME_REMAINING_CHUNKS
        every time it's called. It also handles seconds.

        Args:
            secs_left (int):
                The estimated number of seconds remaining.

        Returns:
            unicode:
            The text containing the time remaining.
        """
        delta = timedelta(seconds=secs_left)
        since = delta.days * 24 * 60 * 60 + delta.seconds

        if since < 60:
            return N_('%d second', '%d seconds') % since

        for i, (seconds, name) in enumerate(self.TIME_REMAINING_CHUNKS):
            count = since // seconds

            if count != 0:
                break

        result = name % count

        if i + 1 < len(self.TIME_REMAINING_CHUNKS):
            seconds2, name2 = self.TIME_REMAINING_CHUNKS[i + 1]
            count2 = (since - (seconds * count)) // seconds2

            if count2 != 0:
                result += ', ' + name2 % count2

        return result



# coding=utf-8

import os
import unittest
from hashlib import md5

import kgb
from django.conf import settings
from djblets.testing.decorators import add_fixtures

from reviewboard.diffviewer.diffutils import patch
from reviewboard.diffviewer.testing.mixins import DiffParserTestingMixin
from reviewboard.scmtools.core import (Branch, Commit, Revision, HEAD,
                                       PRE_CREATION)
from reviewboard.scmtools.errors import SCMError, FileNotFoundError
from reviewboard.scmtools.svn import (
    SVNTool,
    recompute_svn_backend,
    _IDEA_EMPTY,
)
from reviewboard.scmtools.svn.utils import (collapse_svn_keywords,
                                            has_expanded_svn_keywords)
from reviewboard.scmtools.tests.testcases import SCMTestCase
from reviewboard.testing.testcase import TestCase


class _CommonSVNTestCase(DiffParserTestingMixin, kgb.SpyAgency, SCMTestCase):
    """Common unit tests for Subversion.

    This is meant to be subclassed for each backend that wants to run
    the common set of tests.
    """

    backend = None
    backend_name = None
    fixtures = ['test_scmtools']

    ssh_required_system_exes = ['svn', 'svnserve']

    __test__ = False

    @classmethod
    def setUpClass(cls):
        super(_CommonSVNTestCase, cls).setUpClass()

        cls._old_backend_setting = settings.SVNTOOL_BACKENDS
        settings.SVNTOOL_BACKENDS = [cls.backend]
        recompute_svn_backend()

    @classmethod
    def tearDownClass(cls):
        super(_CommonSVNTestCase, cls).tearDownClass()

        settings.SVNTOOL_BACKENDS = cls._old_backend_setting
        recompute_svn_backend()

    def setUp(self):
        super(_CommonSVNTestCase, self).setUp()

        self.svn_repo_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__),
                         '..', 'testdata', 'svn_repo'))
        self.svn_ssh_path = ('svn+ssh://localhost%s'
                             % self.svn_repo_path.replace('\\', '/'))
        self.repository = self.create_repository(
            name='Subversion SVN',
            path='file://%s' % self.svn_repo_path,
            tool_name='Subversion')

        try:
            self.tool = self.repository.get_scmtool()
        except ImportError:
            raise unittest.SkipTest('The %s backend could not be used. A '
                                    'dependency may be missing.'
                                    % self.backend)

        assert self.tool.client.__class__.__module__ == self.backend

    def shortDescription(self):
        desc = super(_CommonSVNTestCase, self).shortDescription()
        desc = desc.replace('<backend>', self.backend_name)

        return desc

    def test_get_repository_info(self):
        """Testing SVN (<backend>) get_repository_info"""
        info = self.tool.get_repository_info()

        self.assertIn('uuid', info)
        self.assertIsInstance(info['uuid'], str)
        self.assertEqual(info['uuid'], '41215d38-f5a5-421f-ba17-e0be11e6c705')

        self.assertIn('root_url', info)
        self.assertIsInstance(info['root_url'], str)
        self.assertEqual(info['root_url'], self.repository.path)

        self.assertIn('url', info)
        self.assertIsInstance(info['url'], str)
        self.assertEqual(info['url'], self.repository.path)

    def test_ssh(self):
        """Testing SVN (<backend>) with a SSH-backed Subversion repository"""
        self._test_ssh(self.svn_ssh_path, 'trunk/doc/misc-docs/Makefile')

    def test_ssh_with_site(self):
        """Testing SVN (<backend>) with a SSH-backed Subversion repository
        with a LocalSite
        """
        self._test_ssh_with_site(self.svn_ssh_path,
                                 'trunk/doc/misc-docs/Makefile')

    def test_get_file(self):
        """Testing SVN (<backend>) get_file"""
        tool = self.tool

        expected = (b'include ../tools/Makefile.base-vars\n'
                    b'NAME = misc-docs\n'
                    b'OUTNAME = svn-misc-docs\n'
                    b'INSTALL_DIR = $(DESTDIR)/usr/share/doc/subversion\n'
                    b'include ../tools/Makefile.base-rules\n')

        # There are 3 versions of this test in order to get 100% coverage of
        # the svn module.
        rev = Revision('2')
        filename = 'trunk/doc/misc-docs/Makefile'

        value = tool.get_file(filename, rev)
        self.assertIsInstance(value, bytes)
        self.assertEqual(value, expected)

        value = tool.get_file('/%s' % filename, rev)
        self.assertIsInstance(value, bytes)
        self.assertEqual(value, expected)

        value = tool.get_file('%s/%s' % (self.repository.path, filename), rev)
        self.assertIsInstance(value, bytes)
        self.assertEqual(value, expected)

        with self.assertRaises(FileNotFoundError):
            tool.get_file('')

    def test_file_exists(self):
        """Testing SVN (<backend>) file_exists"""
        tool = self.tool

        self.assertTrue(tool.file_exists('trunk/doc/misc-docs/Makefile'))
        self.assertFalse(tool.file_exists('trunk/doc/misc-docs/Makefile2'))

        with self.assertRaises(FileNotFoundError):
            tool.get_file('hello', PRE_CREATION)

    def test_get_file_with_special_url_chars(self):
        """Testing SVN (<backend>) get_file with filename containing
        characters that are special in URLs and repository path as a URI
        """
        value = self.tool.get_file('trunk/crazy& ?#.txt', Revision('12'))
        self.assertTrue(isinstance(value, bytes))
        self.assertEqual(value, b'Lots of characters in this one.\n')

    def test_file_exists_with_special_url_chars(self):
        """Testing SVN (<backend>) file_exists with filename containing
        characters that are special in URLs
        """
        self.assertTrue(self.tool.file_exists('trunk/crazy& ?#.txt',
                                              Revision('12')))

        # These should not crash. We'll be testing both file:// URLs
        # (which fail for anything lower than ASCII code 32) and for actual
        # URLs (which support all characters).
        self.assertFalse(self.tool.file_exists('trunk/%s.txt' % ''.join(
            chr(c)
            for c in range(32, 128)
        )))

        self.tool.client.repopath = 'svn+ssh://localhost:0/svn'

        try:
            self.assertFalse(self.tool.file_exists('trunk/%s.txt' % ''.join(
                chr(c)
                for c in range(128)
            )))
        except SCMError:
            # Couldn't connect. Valid result.
            pass

    def test_normalize_path_with_special_chars_and_remote_url(self):
        """Testing SVN (<backend>) normalize_path with special characters
        and remote URL
        """
        client = self.tool.client

        client.repopath = 'svn+ssh://example.com/svn'
        path = client.normalize_path(''.join(
            chr(c)
            for c in range(128)
        ))

        # This URL was generated based on modified code that directly used
        # Subversion's lookup take explicitly, ensuring we're getting the
        # results we want from urllib.quote() and our list of safe characters.
        self.assertEqual(
            path,
            "svn+ssh://example.com/svn/%00%01%02%03%04%05%06%07%08%09%0A"
            "%0B%0C%0D%0E%0F%10%11%12%13%14%15%16%17%18%19%1A%1B%1C%1D%1E"
            "%1F%20!%22%23$%25&'()*+,-./0123456789:%3B%3C=%3E%3F@ABCDEFGH"
            "IJKLMNOPQRSTUVWXYZ%5B%5C%5D%5E_%60abcdefghijklmnopqrstuvwxyz"
            "%7B%7C%7D~%7F")

    def test_normalize_path_with_special_chars_and_file_url(self):
        """Testing SVN (<backend>) normalize_path with special characters
        and local file:// URL
        """
        client = self.tool.client

        client.repopath = 'file:///tmp/svn'
        path = client.normalize_path(''.join(
            chr(c)
            for c in range(32, 128)
        ))

        # This URL was generated based on modified code that directly used
        # Subversion's lookup take explicitly, ensuring we're getting the
        # results we want from urllib.quote() and our list of safe characters.
        self.assertEqual(
            path,
            "file:///tmp/svn/%20!%22%23$%25&'()*+,-./0123456789:%3B%3C=%3E"
            "%3F@ABCDEFGHIJKLMNOPQRSTUVWXYZ%5B%5C%5D%5E_%60abcdefghijklmno"
            "pqrstuvwxyz%7B%7C%7D~%7F")

        # This should provide a reasonable error for each code in 0..32.
        for i in range(32):
            c = chr(i)

            message = (
                'Invalid character code %s found in path %r.'
                % (i, c)
            )

            with self.assertRaisesMessage(SCMError, message):
                client.normalize_path(c)

    def test_normalize_path_with_absolute_repo_path(self):
        """Testing SVN (<backend>) normalize_path with absolute path"""
        client = self.tool.client

        client.repopath = '/var/lib/svn'
        path = '/var/lib/svn/foo/bar'
        self.assertEqual(client.normalize_path(path), path)

        client.repopath = 'svn+ssh://example.com/svn/'
        path = 'svn+ssh://example.com/svn/foo/bar'
        self.assertEqual(client.normalize_path(path), path)

    def test_normalize_path_with_rel_path(self):
        """Testing SVN (<backend>) normalize_path with relative path"""
        client = self.tool.client
        client.repopath = 'svn+ssh://example.com/svn'

        self.assertEqual(client.normalize_path('foo/bar'),
                         'svn+ssh://example.com/svn/foo/bar')
        self.assertEqual(client.normalize_path('/foo/bar'),
                         'svn+ssh://example.com/svn/foo/bar')
        self.assertEqual(client.normalize_path('//foo/bar'),
                         'svn+ssh://example.com/svn/foo/bar')
        self.assertEqual(client.normalize_path('foo&/b ar?/#file#.txt'),
                         'svn+ssh://example.com/svn/foo&/b%20ar%3F/'
                         '%23file%23.txt')

    def test_revision_parsing(self):
        """Testing SVN (<backend>) revision number parsing"""
        parser = self.tool.get_parser(b'')

        self.assertEqual(
            parser.parse_diff_revision(filename=b'',
                                       revision=b'(working copy)'),
            (b'', HEAD))
        self.assertEqual(
            parser.parse_diff_revision(filename=b'',
                                       revision=b'   (revision 0)'),
            (b'', PRE_CREATION))

        self.assertEqual(
            parser.parse_diff_revision(filename=b'',
                                       revision=b'(revision 1)'),
            (b'', b'1'))
        self.assertEqual(
            parser.parse_diff_revision(filename=b'',
                                       revision=b'(revision 23)'),
            (b'', b'23'))

        # Fix for bug 2176
        self.assertEqual(
            parser.parse_diff_revision(filename=b'',
                                       revision=b'\t(revision 4)'),
            (b'', b'4'))

        self.assertEqual(
            parser.parse_diff_revision(
                filename=b'',
                revision=b'2007-06-06 15:32:23 UTC (rev 10958)'),
            (b'', b'10958'))

        # Fix for bug 2632
        self.assertEqual(
            parser.parse_diff_revision(filename=b'',
                                       revision=b'(revision )'),
            (b'', _IDEA_EMPTY))

        with self.assertRaises(SCMError):
            parser.parse_diff_revision(filename=b'',
                                       revision=b'hello')

        # Verify that 'svn diff' localized revision strings parse correctly.
        self.assertEqual(
            parser.parse_diff_revision(
                filename=b'',
                revision='(revisión: 5)'.encode('utf-8')),
            (b'', b'5'))
        self.assertEqual(
            parser.parse_diff_revision(
                filename=b'',
                revision='(リビジョン 6)'.encode('utf-8')),
            (b'', b'6'))
        self.assertEqual(
            parser.parse_diff_revision(
                filename=b'',
                revision='(版本 7)'.encode('utf-8')),
            (b'', b'7'))

    def test_revision_parsing_with_nonexistent(self):
        """Testing SVN (<backend>) revision parsing with "(nonexistent)"
        revision indicator
        """
        parser = self.tool.get_parser(b'')

        # English
        self.assertEqual(
            parser.parse_diff_revision(filename=b'',
                                       revision=b'(nonexistent)'),
            (b'', PRE_CREATION))

        # German
        self.assertEqual(
            parser.parse_diff_revision(filename=b'',
                                       revision=b'(nicht existent)'),
            (b'', PRE_CREATION))

        # Simplified Chinese
        self.assertEqual(
            parser.parse_diff_revision(
                filename=b'',
                revision='(不存在的)'.encode('utf-8')),
            (b'', PRE_CREATION))

    def test_revision_parsing_with_nonexistent_and_branches(self):
        """Testing SVN (<backend>) revision parsing with relocation
        information and nonexistent revision specifier
        """
        parser = self.tool.get_parser(b'')

        self.assertEqual(
            parser.parse_diff_revision(
                filename=b'',
                revision=b'(.../trunk) (nonexistent)'),
            (b'trunk/', PRE_CREATION))

        self.assertEqual(
            parser.parse_diff_revision(
                filename=b'',
                revision=b'(.../branches/branch-1.0)     (nicht existent)'),
            (b'branches/branch-1.0/', PRE_CREATION))

        self.assertEqual(
            parser.parse_diff_revision(
                filename=b'',
                revision='        (.../trunk)     (不存在的)'.encode('utf-8')),
            (b'trunk/', PRE_CREATION))

    def test_interface(self):
        """Testing SVN (<backend>) with basic SVNTool API"""
        self.assertFalse(self.tool.diffs_use_absolute_paths)

        self.assertRaises(NotImplementedError,
                          lambda: self.tool.get_changeset(1))

    def test_binary_diff(self):
        """Testing SVN (<backend>) parsing SVN diff with binary file"""
        diff = (
            b'Index: binfile\n'
            b'============================================================'
            b'=======\n'
            b'Cannot display: file marked as a binary type.\n'
            b'svn:mime-type = application/octet-stream\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'binfile',
            orig_file_details=b'(unknown)',
            modified_filename=b'binfile',
            modified_file_details=b'(working copy)',
            index_header_value=b'binfile',
            binary=True,
            data=diff)

    def test_binary_diff_with_property_change(self):
        """Testing SVN (<backend>) parsing SVN diff with binary file with
        property change
        """
        diff = (
            b'Index: binfile\n'
            b'============================================================'
            b'=======\n'
            b'Cannot display: file marked as a binary type.\n'
            b'svn:mime-type = application/octet-stream\n'
            b'\n'
            b'Property changes on: binfile\n'
            b'____________________________________________________________'
            b'_______\n'
            b'Added: svn:mime-type\n'
            b'## -0,0 +1 ##\n'
            b'+application/octet-stream\n'
            b'\\ No newline at end of property\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'binfile',
            orig_file_details=b'(unknown)',
            modified_filename=b'binfile',
            modified_file_details=b'(working copy)',
            index_header_value=b'binfile',
            binary=True,
            insert_count=1,
            data=diff)

    def test_binary_diff_with_revision(self) -> None:
        """Testing SVN (<backend>) parsing diff with binary file that has
        revision information, generated with --force
        """
        diff = (
            b'Index: binfile\n'
            b'============================================================'
            b'=======\n'
            b'Binary files binfile (revision 3) and binfile (working copy) '
            b'differ\n'
            b'Cannot display: file marked as a binary type.\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'binfile',
            orig_file_details=b'3',
            modified_filename=b'binfile',
            modified_file_details=HEAD,
            index_header_value=b'binfile',
            binary=True,
            insert_count=0,
            data=diff)

    def test_keyword_diff(self):
        """Testing SVN (<backend>) parsing diff with keywords"""
        # 'svn cat' will expand special variables in svn:keywords,
        # but 'svn diff' doesn't expand anything.  This causes the
        # patch to fail if those variables appear in the patch context.
        diff = (b'Index: Makefile\n'
                b'==========================================================='
                b'========\n'
                b'--- Makefile    (revision 4)\n'
                b'+++ Makefile    (working copy)\n'
                b'@@ -1,6 +1,7 @@\n'
                b' # $Id$\n'
                b' # $Rev$\n'
                b' # $Revision::     $\n'
                b'+# foo\n'
                b' include ../tools/Makefile.base-vars\n'
                b' NAME = misc-docs\n'
                b' OUTNAME = svn-misc-docs\n')

        filename = 'trunk/doc/misc-docs/Makefile'
        rev = Revision('4')
        file = self.tool.get_file(filename, rev)
        patch(diff, file, filename)

    def test_unterminated_keyword_diff(self):
        """Testing SVN (<backend>) parsing diff with unterminated keywords"""
        diff = (b'Index: Makefile\n'
                b'==========================================================='
                b'========\n'
                b'--- Makefile    (revision 4)\n'
                b'+++ Makefile    (working copy)\n'
                b'@@ -1,6 +1,7 @@\n'
                b' # $Id$\n'
                b' # $Id:\n'
                b' # $Rev$\n'
                b' # $Revision::     $\n'
                b'+# foo\n'
                b' include ../tools/Makefile.base-vars\n'
                b' NAME = misc-docs\n'
                b' OUTNAME = svn-misc-docs\n')

        filename = 'trunk/doc/misc-docs/Makefile'
        rev = Revision('5')
        file = self.tool.get_file(filename, rev)
        patch(diff, file, filename)

    def test_svn16_property_diff(self):
        """Testing SVN (<backend>) parsing SVN 1.6 diff with property changes
        """
        diff = (
            b'Index:\n'
            b'======================================================'
            b'=============\n'
            b'--- (revision 123)\n'
            b'+++ (working copy)\n'
            b'Property changes on: .\n'
            b'______________________________________________________'
            b'_____________\n'
            b'Modified: reviewboard:url\n'
            b'## -1 +1 ##\n'
            b'-http://reviews.reviewboard.org\n'
            b'+http://reviews.reviewboard.org\n'
            b'Index: binfile\n'
            b'======================================================='
            b'============\nCannot display: file marked as a '
            b'binary type.\nsvn:mime-type = application/octet-stream\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'binfile',
            orig_file_details=b'(unknown)',
            modified_filename=b'binfile',
            modified_file_details=b'(working copy)',
            index_header_value=b'binfile',
            binary=True,
            data=diff)

    def test_svn17_property_diff(self):
        """Testing SVN (<backend>) parsing SVN 1.7+ diff with property changes
        """
        diff = (
            b'Index .:\n'
            b'======================================================'
            b'=============\n'
            b'--- .  (revision 123)\n'
            b'+++ .  (working copy)\n'
            b'\n'
            b'Property changes on: .\n'
            b'______________________________________________________'
            b'_____________\n'
            b'Modified: reviewboard:url\n'
            b'## -0,0 +1,3 ##\n'
            b'-http://reviews.reviewboard.org\n'
            b'+http://reviews.reviewboard.org\n'
            b'Added: myprop\n'
            b'## -0,0 +1 ##\n'
            b'+Property test.\n'
            b'Index: binfile\n'
            b'======================================================='
            b'============\nCannot display: file marked as a '
            b'binary type.\nsvn:mime-type = application/octet-stream\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'binfile',
            orig_file_details=b'(unknown)',
            modified_filename=b'binfile',
            modified_file_details=b'(working copy)',
            index_header_value=b'binfile',
            binary=True,
            data=diff)

    def test_unicode_diff(self):
        """Testing SVN (<backend>) parsing diff with unicode characters"""
        diff = (
            'Index: Filé\n'
            '==========================================================='
            '========\n'
            '--- Filé    (revision 4)\n'
            '+++ Filé    (working copy)\n'
            '@@ -1,6 +1,7 @@\n'
            '+# foó\n'
            ' include ../tools/Makefile.base-vars\n'
            ' NAME = misc-docs\n'
            ' OUTNAME = svn-misc-docs\n'
        ).encode('utf-8')

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename='Filé'.encode('utf-8'),
            orig_file_details=b'4',
            modified_filename='Filé'.encode('utf-8'),
            modified_file_details=HEAD,
            index_header_value='Filé'.encode('utf-8'),
            insert_count=1,
            data=diff)

    def test_diff_with_spaces_in_filenames(self):
        """Testing SVN (<backend>) parsing diff with spaces in filenames"""
        diff = (
            b'Index: File with spaces\n'
            b'==========================================================='
            b'========\n'
            b'--- File with spaces    (revision 4)\n'
            b'+++ File with spaces    (working copy)\n'
            b'@@ -1,6 +1,7 @@\n'
            b'+# foo\n'
            b' include ../tools/Makefile.base-vars\n'
            b' NAME = misc-docs\n'
            b' OUTNAME = svn-misc-docs\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'File with spaces',
            orig_file_details=b'4',
            modified_filename=b'File with spaces',
            modified_file_details=HEAD,
            index_header_value=b'File with spaces',
            insert_count=1,
            data=diff)

    def test_diff_with_added_empty_file(self):
        """Testing parsing SVN diff with added empty file"""
        diff = (
            b'Index: empty-file\t(added)\n'
            b'==========================================================='
            b'========\n'
            b'--- empty-file\t(revision 0)\n'
            b'+++ empty-file\t(revision 0)\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'empty-file',
            orig_file_details=PRE_CREATION,
            modified_filename=b'empty-file',
            modified_file_details=PRE_CREATION,
            index_header_value=b'empty-file\t(added)',
            data=diff)

    def test_diff_with_deleted_empty_file(self):
        """Testing parsing SVN diff with deleted empty file"""
        diff = (
            b'Index: empty-file\t(deleted)\n'
            b'==========================================================='
            b'========\n'
            b'--- empty-file\t(revision 4)\n'
            b'+++ empty-file\t(working copy)\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'empty-file',
            orig_file_details=b'4',
            modified_filename=b'empty-file',
            modified_file_details=HEAD,
            index_header_value=b'empty-file\t(deleted)',
            deleted=True,
            data=diff)

    def test_diff_with_nonexistent_revision_for_dest_file(self):
        """Testing parsing SVN diff with deleted file using "nonexistent"
        destination revision
        """
        diff = (
            b'Index: deleted-file\n'
            b'==========================================================='
            b'========\n'
            b'--- deleted-file\t(revision 4)\n'
            b'+++ deleted-file\t(nonexistent)\n'
            b'@@ -1,2 +0,0 @@\n'
            b'-line 1\n'
            b'-line 2\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'deleted-file',
            orig_file_details=b'4',
            modified_filename=b'deleted-file',
            modified_file_details=PRE_CREATION,
            index_header_value=b'deleted-file',
            deleted=True,
            delete_count=2,
            data=diff)

    def test_idea_diff(self):
        """Testing parsing SVN diff with multi-file diff generated by IDEA
        IDEs
        """
        diff1 = (
            b'Index: path/to/README\n'
            b'IDEA additional info:\n'
            b'Subsystem: org.reviewboard.org.test\n'
            b'<+>ISO-8859-1\n'
            b'=============================================================='
            b'=====\n'
            b'--- path/to/README\t(revision 4)\n'
            b'+++ path/to/README\t(revision )\n'
            b'@@ -1,6 +1,7 @@\n'
            b' #\n'
            b' #\n'
            b' #\n'
            b'+# test\n'
            b' #\n'
            b' #\n'
            b' #\n'
        )
        diff2 = (
            b'Index: path/to/README2\n'
            b'IDEA additional info:\n'
            b'Subsystem: org.reviewboard.org.test\n'
            b'<+>ISO-8859-1\n'
            b'=============================================================='
            b'=====\n'
            b'--- path/to/README2\t(revision 4)\n'
            b'+++ path/to/README2\t(revision )\n'
            b'@@ -1,6 +1,7 @@\n'
            b' #\n'
            b' #\n'
            b' #\n'
            b'+# test\n'
            b' #\n'
            b' #\n'
            b' #\n'
        )
        diff = diff1 + diff2

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 2)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'path/to/README',
            orig_file_details=b'4',
            modified_filename=b'path/to/README',
            modified_file_details=HEAD,
            index_header_value=b'path/to/README',
            insert_count=1,
            data=diff1)

        self.assert_parsed_diff_file(
            parsed_files[1],
            orig_filename=b'path/to/README2',
            orig_file_details=b'4',
            modified_filename=b'path/to/README2',
            modified_file_details=HEAD,
            index_header_value=b'path/to/README2',
            insert_count=1,
            data=diff2)

    def test_get_branches(self):
        """Testing SVN (<backend>) get_branches"""
        branches = self.tool.get_branches()

        self.assertEqual(len(branches), 3)
        self.assertEqual(branches[0], Branch(id='trunk', name='trunk',
                                             commit='12', default=True))
        self.assertEqual(branches[1], Branch(id='branches/branch1',
                                             name='branch1',
                                             commit='7', default=False))
        self.assertEqual(branches[2], Branch(id='top-level-branch',
                                             name='top-level-branch',
                                             commit='10', default=False))

    def test_get_commits(self):
        """Testing SVN (<backend>) get_commits"""
        commits = self.tool.get_commits(start='5')

        self.assertEqual(len(commits), 5)
        self.assertEqual(
            commits[0],
            Commit('chipx86',
                   '5',
                   '2010-05-21T09:33:40.893946',
                   'Add an unterminated keyword for testing bug #1523\n',
                   '4'))

        commits = self.tool.get_commits(start='7')
        self.assertEqual(len(commits), 7)
        self.assertEqual(
            commits[1],
            Commit('david',
                   '6',
                   '2013-06-13T07:43:04.725088',
                   'Add a branches directory',
                   '5'))

    def test_get_commits_with_branch(self):
        """Testing SVN (<backend>) get_commits with branch"""
        commits = self.tool.get_commits(branch='/branches/branch1', start='5')

        self.assertEqual(len(commits), 5)
        self.assertEqual(
            commits[0],
            Commit('chipx86',
                   '5',
                   '2010-05-21T09:33:40.893946',
                   'Add an unterminated keyword for testing bug #1523\n',
                   '4'))

        commits = self.tool.get_commits(branch='/branches/branch1', start='7')
        self.assertEqual(len(commits), 6)
        self.assertEqual(
            commits[0],
            Commit('david',
                   '7',
                   '2013-06-13T07:43:27.259554',
                   'Add a branch',
                   '5'))
        self.assertEqual(
            commits[1],
            Commit('chipx86',
                   '5',
                   '2010-05-21T09:33:40.893946',
                   'Add an unterminated keyword for testing bug #1523\n',
                   '4'))

    def test_get_commits_with_no_date(self):
        """Testing SVN (<backend>) get_commits with no date in commit"""
        self.spy_on(self.tool.client.get_log, op=kgb.SpyOpReturn([
            {
                'author': 'chipx86',
                'message': 'Commit 1',
                'revision': '5',
            },
        ]))

        commits = self.tool.get_commits(start='5')

        self.assertEqual(len(commits), 1)
        self.assertEqual(
            commits[0],
            Commit('chipx86',
                   '5',
                   '',
                   'Commit 1'))

    def test_get_commits_with_exception(self):
        """Testing SVN (<backend>) get_commits with exception"""
        self.spy_on(self.tool.client.communicate_hook,
                    op=kgb.SpyOpRaise(Exception('Bad things happened')))

        with self.assertRaisesMessage(SCMError, 'Bad things happened'):
            self.tool.get_commits(start='5')

    def test_get_change(self):
        """Testing SVN (<backend>) get_change"""
        commit = self.tool.get_change('5')

        self.assertEqual(md5(commit.message.encode('utf-8')).hexdigest(),
                         '928336c082dd756e3f7af4cde4724ebf')
        self.assertEqual(md5(commit.diff).hexdigest(),
                         '56e50374056931c03a333f234fa63375')

    def test_utf8_keywords(self):
        """Testing SVN (<backend>) with UTF-8 files with keywords"""
        self.repository.get_file(path='trunk/utf8-file.txt',
                                 revision='9')

    def test_normalize_patch_with_svn_and_expanded_keywords(self):
        """Testing SVN (<backend>) normalize_patch with expanded keywords"""
        diff = (
            b'Index: Makefile\n'
            b'==========================================================='
            b'========\n'
            b'--- Makefile    (revision 4)\n'
            b'+++ Makefile    (working copy)\n'
            b'@@ -1,6 +1,7 @@\n'
            b' # $Id$\n'
            b' # $Rev: 123$\n'
            b' # $Revision:: 123   $\n'
            b'+# foo\n'
            b' include ../tools/Makefile.base-vars\n'
            b' NAME = misc-docs\n'
            b' OUTNAME = svn-misc-docs\n'
        )

        normalized = self.tool.normalize_patch(
            patch=diff,
            filename='trunk/doc/misc-docs/Makefile',
            revision='4')

        self.assertEqual(
            normalized,
            b'Index: Makefile\n'
            b'==========================================================='
            b'========\n'
            b'--- Makefile    (revision 4)\n'
            b'+++ Makefile    (working copy)\n'
            b'@@ -1,6 +1,7 @@\n'
            b' # $Id$\n'
            b' # $Rev$\n'
            b' # $Revision::       $\n'
            b'+# foo\n'
            b' include ../tools/Makefile.base-vars\n'
            b' NAME = misc-docs\n'
            b' OUTNAME = svn-misc-docs\n')

    def test_normalize_patch_with_svn_and_no_expanded_keywords(self):
        """Testing SVN (<backend>) normalize_patch with no expanded keywords"""
        diff = (
            b'Index: Makefile\n'
            b'==========================================================='
            b'========\n'
            b'--- Makefile    (revision 4)\n'
            b'+++ Makefile    (working copy)\n'
            b'@@ -1,6 +1,7 @@\n'
            b' # $Id$\n'
            b' # $Rev$\n'
            b' # $Revision::    $\n'
            b'+# foo\n'
            b' include ../tools/Makefile.base-vars\n'
            b' NAME = misc-docs\n'
            b' OUTNAME = svn-misc-docs\n'
        )

        normalized = self.tool.normalize_patch(
            patch=diff,
            filename='trunk/doc/misc-docs/Makefile',
            revision='4')

        self.assertEqual(
            normalized,
            b'Index: Makefile\n'
            b'==========================================================='
            b'========\n'
            b'--- Makefile    (revision 4)\n'
            b'+++ Makefile    (working copy)\n'
            b'@@ -1,6 +1,7 @@\n'
            b' # $Id$\n'
            b' # $Rev$\n'
            b' # $Revision::    $\n'
            b'+# foo\n'
            b' include ../tools/Makefile.base-vars\n'
            b' NAME = misc-docs\n'
            b' OUTNAME = svn-misc-docs\n')


class PySVNTests(_CommonSVNTestCase):
    backend = 'reviewboard.scmtools.svn.pysvn'
    backend_name = 'pysvn'

    __test__ = True


class UtilsTests(SCMTestCase):
    """Unit tests for reviewboard.scmtools.svn.utils."""

    def test_collapse_svn_keywords(self):
        """Testing collapse_svn_keywords"""
        keyword_test_data = [
            (b'Id',
             b'/* $Id: test2.c 3 2014-08-04 22:55:09Z david $ */',
             b'/* $Id$ */'),
            (b'id',
             b'/* $Id: test2.c 3 2014-08-04 22:55:09Z david $ */',
             b'/* $Id$ */'),
            (b'id',
             b'/* $id: test2.c 3 2014-08-04 22:55:09Z david $ */',
             b'/* $id$ */'),
            (b'Id',
             b'/* $id: test2.c 3 2014-08-04 22:55:09Z david $ */',
             b'/* $id$ */')
        ]

        for keyword, data, result in keyword_test_data:
            self.assertEqual(collapse_svn_keywords(data, keyword),
                             result)

    def test_has_expanded_svn_keywords(self):
        """Testing has_expanded_svn_keywords"""
        self.assertTrue(has_expanded_svn_keywords(b'.. $ID: 123$ ..'))
        self.assertTrue(has_expanded_svn_keywords(b'.. $id::  123$ ..'))

        self.assertFalse(has_expanded_svn_keywords(b'.. $Id::   $ ..'))
        self.assertFalse(has_expanded_svn_keywords(b'.. $Id$ ..'))
        self.assertFalse(has_expanded_svn_keywords(b'.. $Id ..'))
        self.assertFalse(has_expanded_svn_keywords(b'.. $Id Here$ ..'))


class SVNAuthFormTests(TestCase):
    """Unit tests for SVNTool's authentication form."""

    def test_fields(self):
        """Testing SVNTool authentication form fields"""
        form = SVNTool.create_auth_form()

        self.assertEqual(list(form.fields), ['username', 'password'])
        self.assertEqual(form['username'].help_text, '')
        self.assertEqual(form['username'].label, 'Username')
        self.assertEqual(form['password'].help_text, '')
        self.assertEqual(form['password'].label, 'Password')

    @add_fixtures(['test_scmtools'])
    def test_load(self):
        """Tetting SVNTool authentication form load"""
        repository = self.create_repository(
            tool_name='Subversion',
            username='test-user',
            password='test-pass')

        form = SVNTool.create_auth_form(repository=repository)
        form.load()

        self.assertEqual(form['username'].value(), 'test-user')
        self.assertEqual(form['password'].value(), 'test-pass')

    @add_fixtures(['test_scmtools'])
    def test_save(self):
        """Tetting SVNTool authentication form save"""
        repository = self.create_repository(tool_name='Subversion')

        form = SVNTool.create_auth_form(
            repository=repository,
            data={
                'username': 'test-user',
                'password': 'test-pass',
            })
        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(repository.username, 'test-user')
        self.assertEqual(repository.password, 'test-pass')


class SVNRepositoryFormTests(TestCase):
    """Unit tests for SVNTool's repository form."""

    def test_fields(self) -> None:
        """Testing SVNTool repository form fields"""
        form = SVNTool.create_repository_form()

        self.assertEqual(list(form.fields), ['path', 'mirror_path'])
        self.assertEqual(form['path'].help_text,
                         'The path to the repository. This will generally be '
                         'the URL you would use to check out the repository.')
        self.assertEqual(form['path'].label, 'Path')
        self.assertEqual(form['mirror_path'].help_text, '')
        self.assertEqual(form['mirror_path'].label, 'Mirror Path')

    @add_fixtures(['test_scmtools'])
    def test_load(self) -> None:
        """Tetting SVNTool repository form load"""
        repository = self.create_repository(
            tool_name='Subversion',
            path='https://svn.example.com/',
            mirror_path='https://svn.mirror.example.com')

        form = SVNTool.create_repository_form(repository=repository)
        form.load()

        self.assertEqual(form['path'].value(), 'https://svn.example.com/')
        self.assertEqual(form['mirror_path'].value(),
                         'https://svn.mirror.example.com')

    @add_fixtures(['test_scmtools'])
    def test_save(self) -> None:
        """Tetting SVNTool repository form save"""
        repository = self.create_repository(tool_name='Subversion')

        form = SVNTool.create_repository_form(
            repository=repository,
            data={
                'path': 'https://svn.example.com/',
                'mirror_path': 'https://svn.mirror.example.com',
            })
        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(repository.path, 'https://svn.example.com/')
        self.assertEqual(repository.mirror_path,
                         'https://svn.mirror.example.com')

    @add_fixtures(['test_scmtools'])
    def test_save_with_file_url(self) -> None:
        """Tetting SVNTool repository form save with file:// URL"""
        repository = self.create_repository(tool_name='Subversion')

        form = SVNTool.create_repository_form(
            repository=repository,
            data={
                'path': 'file:///opt/svnrepo/',
            })
        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(repository.path, 'file:///opt/svnrepo/')

    @add_fixtures(['test_scmtools'])
    def test_save_with_bare_file_path(self) -> None:
        """Testing SVNTool repository form save with raw path instead of
        URL
        """
        repository = self.create_repository(tool_name='Subversion')

        form = SVNTool.create_repository_form(
            repository=repository,
            data={
                'path': '/opt/svnrepo/',
            })
        self.assertFalse(form.is_valid())
        form.full_clean()

        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['path'], [
            'The path to the SVN repository must be a URL. To specify a local '
            'repository, use a file:// URL.',
        ])



"""Unit tests for reviewboard.site.models.LocalSite."""

from django.contrib.auth.models import User

from reviewboard.site.models import LocalSite
from reviewboard.testing.testcase import TestCase


class LocalSiteTests(TestCase):
    """Unit tests for LocalSite."""

    fixtures = ['test_users', 'test_site']

    def test_access(self):
        """Test LocalSite.is_accessible_by"""
        doc = User.objects.get(username="doc")
        dopey = User.objects.get(username="dopey")
        site = LocalSite.objects.get(name="local-site-1")

        self.assertTrue(site.is_accessible_by(doc))
        self.assertFalse(site.is_accessible_by(dopey))

    def test_access_with_public(self):
        """Test LocalSite.is_accessible_by with public LocalSites"""
        doc = User.objects.get(username="doc")
        dopey = User.objects.get(username="dopey")
        site = LocalSite.objects.get(name="local-site-1")
        site.public = True

        self.assertTrue(site.is_accessible_by(doc))
        self.assertTrue(site.is_accessible_by(dopey))


class PermissionTests(TestCase):
    """Unit tests for Django-provided permissions for LocalSite."""

    fixtures = ['test_users', 'test_site']

    def setUp(self):
        super().setUp()

        self.user = User.objects.get(username='doc')
        self.assertFalse(self.user.is_superuser)

        self.local_site = LocalSite.objects.get(name=self.local_site_name)
        self.local_site.admins.add(self.user)

    def test_assigned_permissions(self):
        """Testing LocalSite assigned admin permissions"""
        self.assertTrue(self.user.has_perm(
            'hostingsvcs.change_hostingserviceaccount', self.local_site))
        self.assertTrue(self.user.has_perm(
            'hostingsvcs.create_hostingserviceaccount', self.local_site))
        self.assertTrue(self.user.has_perm(
            'reviews.can_change_status', self.local_site))
        self.assertTrue(self.user.has_perm(
            'reviews.can_edit_reviewrequest', self.local_site))
        self.assertTrue(self.user.has_perm(
            'reviews.can_submit_as_another_user', self.local_site))
        self.assertTrue(self.user.has_perm(
            'reviews.change_default_reviewer', self.local_site))
        self.assertTrue(self.user.has_perm(
            'reviews.add_group', self.local_site))
        self.assertTrue(self.user.has_perm(
            'reviews.change_group', self.local_site))
        self.assertTrue(self.user.has_perm(
            'reviews.delete_file', self.local_site))
        self.assertTrue(self.user.has_perm(
            'reviews.delete_screenshot', self.local_site))
        self.assertTrue(self.user.has_perm(
            'scmtools.add_repository', self.local_site))
        self.assertTrue(self.user.has_perm(
            'scmtools.change_repository', self.local_site))

    def test_invalid_permissions(self):
        """Testing LocalSite invalid admin permissions"""
        self.assertFalse(self.user.has_perm(
            'reviews.delete_reviewrequest', self.local_site))
        self.assertFalse(self.user.has_perm(
            'dummy.permission', self.local_site))



"""Unit tests for reviewboard.actions.renderers.MenuItemActionRenderer.

Version Added:
    8.0
"""

from __future__ import annotations

from django.template import Context
from django.utils.safestring import SafeString

from reviewboard.actions.renderers import MenuItemActionRenderer
from reviewboard.actions.tests.base import (TestActionsRegistry,
                                            TestMenuItemAction)
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.testing import TestCase


class MenuItemActionRendererTests(TestCase):
    """Unit tests for MenuItemActionRenderer.

    Version Added:
        8.0
    """

    def test_render(self) -> None:
        """Testing MenuItemActionRenderer.render"""
        action = TestMenuItemAction()
        placement = action.get_placement('review-request')

        registry = TestActionsRegistry()
        registry.register(action)

        renderer = MenuItemActionRenderer(action=action,
                                          placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = renderer.render(request=request,
                               context=context)

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <a id="action-review-request-menu-item-action"
               role="button"
               href="#">
             <span class="my-icon"/>
             Menu Item Action 1
            </a>
            """)

    def test_render_with_action_template(self) -> None:
        """Testing MenuItemActionRenderer.render with action.template_name set
        """
        class MyAction(TestMenuItemAction):
            template_name = 'actions/button_action.html'

        # We already test for the deprecation warning message in the action
        # tests. This just suppresses warning output for the test run.
        with self.assertWarns(RemovedInReviewBoard10_0Warning):
            action = MyAction()

        placement = action.get_placement('review-request')

        registry = TestActionsRegistry()
        registry.register(action)

        renderer = MenuItemActionRenderer(action=action,
                                          placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = renderer.render(request=request,
                               context=context)

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <button class="ink-c-button"
                    id="action-review-request-menu-item-action"
                    type="button"
                    aria-label="Verbose Menu Item Action 1"
                    title="Menu Item 1 description.">
             <span class="ink-c-button__icon my-icon"/>
             <label class="ink-c-button__label">
              Menu Item Action 1
             </label>
            </button>
            """)

    def test_render_js(self) -> None:
        """Testing MenuItemActionRenderer.render_js"""
        action = TestMenuItemAction()
        placement = action.get_placement('review-request')

        registry = TestActionsRegistry()
        registry.register(action)

        renderer = MenuItemActionRenderer(action=action,
                                          placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        js = renderer.render_js(request=request,
                                context=context)

        self.assertIsInstance(js, SafeString)
        self.assertHTMLEqual(
            js,
            """
            page.addActionView(new RB.Actions.MenuItemActionView({
                "attachmentPointID": "review-request",
                el: $('#action-review-request-menu-item-action'),
                model: page.getAction("menu-item-action"),
            }));
            """)
